@@ -2,12 +2,11 @@
     Title:     Write out a database as an ELF object file
     Author:    David Matthews.
 
-    Copyright (c) 2006-7, 2011 David C. J. Matthews
+    Copyright (c) 2006-7, 2011, 2016 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -115,6 +114,7 @@
 #include "run_time.h"
 #include "version.h"
 #include "polystring.h"
+#include "timing.h"
 
 #define sym_last_local_sym sym_data_section
 
@@ -134,6 +134,14 @@ enum {
     // Finally the symbol table
 };
 
+// Add an external reference to the RTS
+void ELFExport::addExternalReference(void *relocAddr, const char *name)
+{
+    externTable.makeEntry(name);
+    // The symbol is added after the memory table entries and poly_exports
+    writeRelocation(0, relocAddr, symbolNum++);
+}
+
 // Generate the address relative to the start of the segment.
 void ELFExport::setRelocationAddress(void *p, ElfXX_Addr *reloc)
 {
@@ -148,13 +156,25 @@ PolyWord ELFExport::createRelocation(PolyWord p, void *relocAddr)
     void *addr = p.AsAddress();
     unsigned addrArea = findArea(addr);
     POLYUNSIGNED offset = (char*)addr - (char*)memTable[addrArea].mtAddr;
+    return writeRelocation(offset, relocAddr, AreaToSym(addrArea));
+}
 
+PolyWord ELFExport::writeRelocation(POLYUNSIGNED offset, void *relocAddr, unsigned symbolNum)
+{
     if (useRela)
     {
         ElfXX_Rela reloc;
         // Set the offset within the section we're scanning.
         setRelocationAddress(relocAddr, &reloc.r_offset);
-        reloc.r_info = ELFXX_R_INFO(AreaToSym(addrArea), directReloc);
+#ifdef HOSTARCHITECTURE_MIPS64
+        reloc.r_sym = symbolNum;
+        reloc.r_ssym = 0;
+        reloc.r_type = directReloc;
+        reloc.r_type2 = 0;
+        reloc.r_type3 = 0;
+#else
+        reloc.r_info = ELFXX_R_INFO(symbolNum, directReloc);
+#endif
         reloc.r_addend = offset;
         fwrite(&reloc, sizeof(reloc), 1, exportFile);
         relocationCount++;
@@ -163,7 +183,15 @@ PolyWord ELFExport::createRelocation(PolyWord p, void *relocAddr)
     else {
         ElfXX_Rel reloc;
         setRelocationAddress(relocAddr, &reloc.r_offset);
-        reloc.r_info = ELFXX_R_INFO(AreaToSym(addrArea), directReloc);
+#ifdef HOSTARCHITECTURE_MIPS64
+        reloc.r_sym = symbolNum;
+        reloc.r_ssym = 0;
+        reloc.r_type = directReloc;
+        reloc.r_type2 = 0;
+        reloc.r_type3 = 0;
+#else
+        reloc.r_info = ELFXX_R_INFO(symbolNum, directReloc);
+#endif
         fwrite(&reloc, sizeof(reloc), 1, exportFile);
         relocationCount++;
         return PolyWord::FromUnsigned(offset);
@@ -173,7 +201,7 @@ PolyWord ELFExport::createRelocation(PolyWord p, void *relocAddr)
 /* This is called for each constant within the code. 
    Print a relocation entry for the word and return a value that means
    that the offset is saved in original word. */
-void ELFExport::ScanConstant(byte *addr, ScanRelocationKind code)
+void ELFExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
     PolyWord p = GetConstantValue(addr, code);
 
@@ -204,7 +232,8 @@ void ELFExport::ScanConstant(byte *addr, ScanRelocationKind code)
             }
         }
         break;
-#if(defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X86_64))
+#if(defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X86_64) || \
+        defined(HOSTARCHITECTURE_X32))
 #ifdef HOSTARCHITECTURE_X86
 #define R_PC_RELATIVE R_386_PC32
 #else
@@ -267,15 +296,13 @@ void ELFExport::writeSymbol(const char *symbolName, long value, long size, int b
     symbol.st_other = 0;
     symbol.st_shndx = section;
     fwrite(&symbol, sizeof(symbol), 1, exportFile);
-    symbolCount++;
 }
 
 // Set the file alignment.
 void ELFExport::alignFile(int align)
 {
-    char pad[32]; // Maximum alignment
+    char pad[32] = {0}; // Maximum alignment
     int offset = ftell(exportFile);
-    memset(pad, 0, sizeof(pad));
     if ((offset % align) == 0) return;
     fwrite(&pad, align - (offset % align), 1, exportFile);
 }
@@ -285,7 +312,15 @@ void ELFExport::createStructsRelocation(unsigned sym, POLYUNSIGNED offset, POLYS
     if (useRela)
     {
         ElfXX_Rela reloc;
+#ifdef HOSTARCHITECTURE_MIPS64
+        reloc.r_sym = sym;
+        reloc.r_ssym = 0;
+        reloc.r_type = directReloc;
+        reloc.r_type2 = 0;
+        reloc.r_type3 = 0;
+#else
         reloc.r_info = ELFXX_R_INFO(sym, directReloc);
+#endif
         reloc.r_offset = offset;
         reloc.r_addend = addend;
         fwrite(&reloc, sizeof(reloc), 1, exportFile);
@@ -294,7 +329,15 @@ void ELFExport::createStructsRelocation(unsigned sym, POLYUNSIGNED offset, POLYS
     else
     {
         ElfXX_Rel reloc;
+#ifdef HOSTARCHITECTURE_MIPS64
+        reloc.r_sym = sym;
+        reloc.r_ssym = 0;
+        reloc.r_type = directReloc;
+        reloc.r_type2 = 0;
+        reloc.r_type3 = 0;
+#else
         reloc.r_info = ELFXX_R_INFO(sym, directReloc);
+#endif
         reloc.r_offset = offset;
         fwrite(&reloc, sizeof(reloc), 1, exportFile);
         relocationCount++;
@@ -306,11 +349,14 @@ void ELFExport::exportStore(void)
     PolyWord    *p;
     ElfXX_Ehdr fhdr;
     ElfXX_Shdr *sections = 0;
-    unsigned numSections = 6 + 2*memTableEntries;
+    unsigned numSections = 6 + 2*memTableEntries /*- 1*/;
     // The symbol table comes at the end.
-    unsigned sect_symtab = sect_data + 2*memTableEntries + 2;
+    unsigned sect_symtab = sect_data + 2*memTableEntries + 2/* - 1*/;
     
     unsigned i;
+    
+    // External symbols start after the memory table entries and "poly_exports".
+    symbolNum = EXTRA_SYMBOLS+memTableEntries+1;
 
     // Both the string tables have an initial null entry.
     symStrings.makeEntry("");
@@ -348,6 +394,18 @@ void ELFExport::exportStore(void)
     fhdr.e_machine = EM_PPC64;
     directReloc = R_PPC64_ADDR64;
     useRela = true;
+#elif defined(HOSTARCHITECTURE_S390)
+    fhdr.e_machine = EM_S390;
+    directReloc = R_390_32;
+    useRela = true;
+#elif defined(HOSTARCHITECTURE_S390X)
+    fhdr.e_machine = EM_S390;
+    directReloc = R_390_64;
+    useRela = true;
+#elif defined(HOSTARCHITECTURE_SH)
+    fhdr.e_machine = EM_SH;
+    directReloc = R_SH_DIR32;
+    useRela = true;
 #elif defined(HOSTARCHITECTURE_SPARC)
     fhdr.e_machine = EM_SPARC;
     directReloc = R_SPARC_32;
@@ -356,12 +414,23 @@ void ELFExport::exportStore(void)
        though, it adds the value in the location being relocated (as with ELF32_Rel
        relocations) as well as the addend. To be safe, whenever we use an ELF32_Rela
        relocation we always zero the location to be relocated. */
+#elif defined(HOSTARCHITECTURE_SPARC64)
+    fhdr.e_machine = EM_SPARCV9;
+    directReloc = R_SPARC_64;
+    /* Use the most relaxed memory model. At link time, the most restrictive one is
+       chosen, so it does no harm to be as permissive as possible here. */
+    fhdr.e_flags = EF_SPARCV9_RMO;
+    useRela = true;
 #elif defined(HOSTARCHITECTURE_X86_64)
     /* It seems Solaris/X86-64 only supports ELF64_Rela relocations.  It appears that
        Linux will support either so we now use Rela on X86-64. */
     fhdr.e_machine = EM_X86_64;
     directReloc = R_X86_64_64;
     useRela = true;
+#elif defined(HOSTARCHITECTURE_X32)
+    fhdr.e_machine = EM_X86_64;
+    directReloc = R_X86_64_32;
+    useRela = true;
 #elif defined(HOSTARCHITECTURE_ARM)
 #ifndef EF_ARM_EABI_VER4 
 #define EF_ARM_EABI_VER4     0x04000000
@@ -373,6 +442,12 @@ void ELFExport::exportStore(void)
     directReloc = R_ARM_ABS32;
     useRela = false;
     fhdr.e_flags = EF_ARM_EABI_VER4;
+#elif defined(HOSTARCHITECTURE_HPPA)
+    fhdr.e_ident[EI_OSABI] = ELFOSABI_HPUX;
+    fhdr.e_machine = EM_PARISC;
+    directReloc = R_PARISC_DIR32;
+    fhdr.e_flags = EFA_PARISC_1_0;
+    useRela = true;
 #elif defined(HOSTARCHITECTURE_IA64)
     fhdr.e_machine = EM_IA_64;
     directReloc = R_IA64_DIR64LSB;
@@ -382,9 +457,28 @@ void ELFExport::exportStore(void)
     fhdr.e_machine = EM_AARCH64;
     directReloc = R_AARCH64_ABS64;
     useRela = true;
+#elif defined(HOSTARCHITECTURE_M68K)
+    fhdr.e_machine = EM_68K;
+    directReloc = R_68K_32;
+    useRela = true;
 #elif defined(HOSTARCHITECTURE_MIPS)
     fhdr.e_machine = EM_MIPS;
     directReloc = R_MIPS_32;
+#ifdef __PIC__
+    fhdr.e_flags = EF_MIPS_CPIC;
+#endif
+    useRela = true;
+#elif defined(HOSTARCHITECTURE_MIPS64)
+    fhdr.e_machine = EM_MIPS;
+    directReloc = R_MIPS_64;
+    fhdr.e_flags = EF_MIPS_ARCH_64;
+#ifdef __PIC__
+    fhdr.e_flags |= EF_MIPS_CPIC;
+#endif
+    useRela = true;
+#elif defined(HOSTARCHITECTURE_ALPHA)
+    fhdr.e_machine = EM_ALPHA;
+    directReloc = R_ALPHA_REFQUAD;
     useRela = true;
 #else
 #error "No support for exporting on this architecture"
@@ -424,14 +518,15 @@ void ELFExport::exportStore(void)
     unsigned long dataRelName = makeStringTableEntry(useRela ? ".rela.data" : ".rel.data", &sectionStrings);
     unsigned long textName = makeStringTableEntry(".text", &sectionStrings);
     unsigned long textRelName = makeStringTableEntry(useRela ? ".rela.text" : ".rel.text", &sectionStrings);
+    unsigned long rodataName = makeStringTableEntry(".rodata", &sectionStrings);
+    unsigned long rodataRelName = makeStringTableEntry(useRela ? ".rela.rodata" : ".rel.rodata", &sectionStrings);
 
     // Main data sections.  Each one has a relocation section.
     for (i=0; i < memTableEntries; i++)
     {
         unsigned s = sect_data + i*2;
-        sections[s].sh_type = SHT_PROGBITS;
         sections[s].sh_addralign = 8; // 8-byte alignment
-
+        sections[s].sh_type = SHT_PROGBITS;
         if (memTable[i].mtFlags & MTF_WRITEABLE)
         {
             // Mutable areas
@@ -440,13 +535,20 @@ void ELFExport::exportStore(void)
             sections[s].sh_flags = SHF_WRITE | SHF_ALLOC;
             sections[s+1].sh_name = dataRelName; // Name of relocation section
         }
-        else
+        else if (memTable[i].mtFlags & MTF_EXECUTABLE)
         {
-            // Immutable areas are marked as executable.
+            // Code areas are marked as executable.
             sections[s].sh_name = textName;
             sections[s].sh_flags = SHF_ALLOC | SHF_EXECINSTR;
             sections[s+1].sh_name = textRelName; // Name of relocation section
         }
+        else
+        {
+            // Non-code immutable areas
+            sections[s].sh_name = rodataName;
+            sections[s].sh_flags = SHF_ALLOC;
+            sections[s+1].sh_name = rodataRelName; // Name of relocation section
+        }
         // sections[s].sh_size is set later
         // sections[s].sh_offset is set later.
         // sections[s].sh_size is set later.
@@ -462,7 +564,7 @@ void ELFExport::exportStore(void)
     }
 
     // Table data - Poly tables that describe the memory layout.
-    unsigned sect_table_data = sect_data+2*memTableEntries;
+    unsigned sect_table_data = sect_data + 2*memTableEntries;
 
     sections[sect_table_data].sh_name = dataName;
     sections[sect_table_data].sh_type = SHT_PROGBITS;
@@ -486,86 +588,28 @@ void ELFExport::exportStore(void)
     // sections[sect_symtab].sh_size is set later
     // sections[sect_symtab].sh_offset is set later
 
-    // First the symbol table.
-    alignFile(sections[sect_symtab].sh_addralign);
-    sections[sect_symtab].sh_offset = ftell(exportFile);
-    writeSymbol("", 0, 0, 0, 0, 0); // Initial symbol
-    // Write the local symbols first.
-    writeSymbol("", 0, 0, STB_LOCAL, STT_SECTION, sect_data); // .data section
-
-    // Create symbols for the address areas.  AreaToSym assumes these come first.
-    for (i = 0; i < memTableEntries; i++)
-    {
-        if (i == ioMemEntry)
-            writeSymbol("ioarea", 0, 0, STB_LOCAL, STT_OBJECT, sect_data+i*2);
-        else {
-            char buff[50];
-            sprintf(buff, "area%1u", i);
-            writeSymbol(buff, 0, 0, STB_LOCAL, STT_OBJECT, sect_data+i*2);
-        }
-    }
-
-    // Extra symbols to help debugging.
-    for (i = 0; i < memTableEntries; i++)
-    {
-        if (i != ioMemEntry)
-        {
-            char buff[50];
-            // Write the names of the functions as local symbols.  This isn't necessary
-            // but it makes debugging easier since the function names appear in gdb.
-            char *start = (char*)memTable[i].mtAddr;
-            char *end = start + memTable[i].mtLength;
-            for (p = (PolyWord*)start; p < (PolyWord*)end; )
-            {
-                p++;
-                PolyObject *obj = (PolyObject*)p;
-                POLYUNSIGNED length = obj->Length();
-                if (length != 0 && obj->IsCodeObject())
-                {
-                    PolyWord *name = obj->ConstPtrForCode();
-                    // Copy as much of the name as will fit and ignore any extra.
-                    // Do we need to worry about duplicates?
-                    (void)Poly_string_to_C(*name, buff, sizeof(buff));
-                    writeSymbol(buff, ((char*)p - start), 0, STB_LOCAL, STT_OBJECT, sect_data+i*2);
-                }
-                p += length;
-            }
-        }
-    }
-
-    // Global symbols - Just one: poly_exports
-    writeSymbol("poly_exports", 0, 
-        sizeof(exportDescription)+sizeof(memoryTableEntry)*memTableEntries,
-        STB_GLOBAL, STT_OBJECT, sect_table_data);
-
-    sections[sect_symtab].sh_info = symbolCount-1; // One more than last local sym
-    sections[sect_symtab].sh_size = sizeof(ElfXX_Sym) * symbolCount;
-
     // Write the relocations.
 
     for (i = 0; i < memTableEntries; i++)
     {
-        unsigned relocSection = sect_data+i*2+1;
+        unsigned relocSection = sect_data + i*2 + 1;
         alignFile(sections[relocSection].sh_addralign);
         sections[relocSection].sh_offset = ftell(exportFile);
         relocationCount = 0;
-        if (i != ioMemEntry) // Don't relocate the IO area
+        // Create the relocation table and turn all addresses into offsets.
+        char *start = (char*)memTable[i].mtAddr;
+        char *end = start + memTable[i].mtLength;
+        for (p = (PolyWord*)start; p < (PolyWord*)end; )
         {
-            // Create the relocation table and turn all addresses into offsets.
-            char *start = (char*)memTable[i].mtAddr;
-            char *end = start + memTable[i].mtLength;
-            for (p = (PolyWord*)start; p < (PolyWord*)end; )
-            {
-                p++;
-                PolyObject *obj = (PolyObject*)p;
-                POLYUNSIGNED length = obj->Length();
-                // Update any constants before processing the object
-                // We need that for relative jumps/calls in X86/64.
-                if (length != 0 && obj->IsCodeObject())
-                    machineDependent->ScanConstantsWithinCode(obj, this);
-                relocateObject(obj);
-                p += length;
-            }
+            p++;
+            PolyObject *obj = (PolyObject*)p;
+            POLYUNSIGNED length = obj->Length();
+            // Update any constants before processing the object
+            // We need that for relative jumps/calls in X86/64.
+            if (length != 0 && obj->IsCodeObject())
+                machineDependent->ScanConstantsWithinCode(obj, this);
+            relocateObject(obj);
+            p += length;
         }
         sections[relocSection].sh_size =
             relocationCount * (useRela ? sizeof(ElfXX_Rela) : sizeof(ElfXX_Rel));
@@ -583,7 +627,7 @@ void ELFExport::exportStore(void)
     // Address of "memTable" within "exports". We can't use createRelocation because
     // the position of the relocation is not in either the mutable or the immutable area.
     POLYSIGNED memTableOffset = (POLYSIGNED)sizeof(exportDescription); // It follows immediately after this.
-    createStructsRelocation(symbolCount-1 /* Last symbol */, offsetof(exportDescription, memTable), memTableOffset);
+    createStructsRelocation(AreaToSym(memTableEntries), offsetof(exportDescription, memTable), memTableOffset);
 
     // Address of "rootFunction" within "exports"
     unsigned rootAddrArea = findArea(rootFunction);
@@ -601,15 +645,41 @@ void ELFExport::exportStore(void)
     sections[sect_table_data+1].sh_size =
         relocationCount * (useRela ? sizeof(ElfXX_Rela) : sizeof(ElfXX_Rel));
 
-    // Now the binary data.
+    // Now the symbol table.
+    alignFile(sections[sect_symtab].sh_addralign);
+    sections[sect_symtab].sh_offset = ftell(exportFile);
+    writeSymbol("", 0, 0, 0, 0, 0); // Initial symbol
+    // Write the local symbols first.
+    writeSymbol("", 0, 0, STB_LOCAL, STT_SECTION, sect_data); // .data section
+
+    // Create symbols for the address areas.  AreaToSym assumes these come first.
+    for (i = 0; i < memTableEntries; i++)
+    {
+        unsigned s = sect_data + i*2;
+        char buff[50];
+        sprintf(buff, "area%1u", i);
+        writeSymbol(buff, 0, 0, STB_LOCAL, STT_OBJECT, s);
+    }
+
+    // Global symbols - Exported symbol for table.
+    writeSymbol("poly_exports", 0, 
+        sizeof(exportDescription)+sizeof(memoryTableEntry)*memTableEntries,
+        STB_GLOBAL, STT_OBJECT, sect_table_data);
+ 
+    // External references
+    for (unsigned i = 0; i < externTable.stringSize; i += (unsigned)strlen(externTable.strings+i) + 1)
+        writeSymbol(externTable.strings+i, 0, 0, STB_GLOBAL, STT_FUNC, SHN_UNDEF);
+
+    sections[sect_symtab].sh_info = EXTRA_SYMBOLS+memTableEntries; // One more than last local sym
+    sections[sect_symtab].sh_size = sizeof(ElfXX_Sym) * symbolNum;
 
     // Now the binary data.
     for (i = 0; i < memTableEntries; i++)
     {
-        unsigned dataSection = sect_data+i*2;
+        unsigned dataSection = sect_data + i*2;
+        sections[dataSection].sh_size = memTable[i].mtLength;
         alignFile(sections[dataSection].sh_addralign);
         sections[dataSection].sh_offset = ftell(exportFile);
-        sections[dataSection].sh_size = memTable[i].mtLength;
         fwrite(memTable[i].mtAddr, 1, memTable[i].mtLength, exportFile);
     }
 
@@ -618,13 +688,11 @@ void ELFExport::exportStore(void)
     exports.structLength = sizeof(exportDescription);
     exports.memTableSize = sizeof(memoryTableEntry);
     exports.memTableEntries = memTableEntries;
-    exports.ioIndex = 0; // The io entry is the first in the memory table
     exports.memTable = useRela ? 0 : (memoryTableEntry *)memTableOffset;
     // Set the value to be the offset relative to the base of the area.  We have set a relocation
     // already which will add the base of the area.
     exports.rootFunction = useRela ? 0 : (void*)rootOffset;
-    exports.timeStamp = time(NULL);
-    exports.ioSpacing = ioSpacing;
+    exports.timeStamp = getBuildTime();
     exports.architecture = machineDependent->MachineArchitecture();
     exports.rtsVersion = POLY_version_number;
 
