@@ -1,7 +1,7 @@
 /*
     Title:  exporter.cpp - Export a function as an object or C file
 
-    Copyright (c) 2006-7, 2015 David C.J. Matthews
+    Copyright (c) 2006-7, 2015, 2016 David C.J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -71,6 +71,7 @@
 #include "memmgr.h"
 #include "processes.h" // For IO_SPACING
 #include "sys.h" // For EXC_Fail
+#include "rtsentry.h"
 
 #include "pexport.h"
 
@@ -82,10 +83,19 @@
 #include "machoexport.h"
 #endif
 
-#if(!defined(MAXPATHLEN) && defined(MAX_PATH))
-#define MAXPATHLEN MAX_PATH
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+#define NOMEMORY ERROR_NOT_ENOUGH_MEMORY
+#define ERRORNUMBER _doserrno
+#else
+#define NOMEMORY ENOMEM
+#define ERRORNUMBER errno
 #endif
 
+extern "C" {
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyExport(PolyObject *threadId, PolyWord fileName, PolyWord root);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyExportPortable(PolyObject *threadId, PolyWord fileName, PolyWord root);
+}
+
 /*
 To export the function and everything reachable from it we need to copy
 all the objects into a new area.  We leave tombstones in the original
@@ -102,7 +112,7 @@ GraveYard::~GraveYard()
 
 CopyScan::CopyScan(unsigned h/*=0*/): hierarchy(h)
 {
-    defaultImmSize = defaultMutSize = 0;
+    defaultImmSize = defaultMutSize = defaultCodeSize = 0;
     defaultNoOverSize = 4096; // This can be small.
     tombs = 0;
     graveYard = 0;
@@ -110,7 +120,7 @@ CopyScan::CopyScan(unsigned h/*=0*/): hierarchy(h)
 
 void CopyScan::initialise(bool isExport/*=true*/)
 {
-    ASSERT(gMem.neSpaces == 0);
+    ASSERT(gMem.eSpaces.size() == 0);
     // Set the space sizes to a proportion of the space currently in use.
     // Computing these sizes is not obvious because CopyScan is used both
     // for export and for saved states.  For saved states in particular we
@@ -119,20 +129,22 @@ void CopyScan::initialise(bool isExport/*=true*/)
     // to waste memory.
     if (hierarchy == 0)
     {
-        graveYard = new GraveYard[gMem.npSpaces];
+        graveYard = new GraveYard[gMem.pSpaces.size()];
         if (graveYard == 0)
             throw MemoryException();
     }
-    unsigned i;
-    for (i = 0; i < gMem.npSpaces; i++)
+
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
     {
-        PermanentMemSpace *space = gMem.pSpaces[i];
+        PermanentMemSpace *space = *i;
         if (space->hierarchy >= hierarchy) {
             // Include this if we're exporting (hierarchy=0) or if we're saving a state
             // and will include this in the new state.
             POLYUNSIGNED size = (space->top-space->bottom)/4;
             if (space->isMutable)
                 defaultMutSize += size;
+            else if (space->isCode)
+                defaultCodeSize += size;
             else
                 defaultImmSize += size;
             if (space->hierarchy == 0 && ! space->isMutable)
@@ -147,9 +159,9 @@ void CopyScan::initialise(bool isExport/*=true*/)
             }
         }
     }
-    for (i = 0; i < gMem.nlSpaces; i++)
+    for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *space = gMem.lSpaces[i];
+        LocalMemSpace *space = *i;
         POLYUNSIGNED size = space->allocatedSpace();
         // It looks as though the mutable size generally gets
         // overestimated while the immutable size is correct.
@@ -158,17 +170,25 @@ void CopyScan::initialise(bool isExport/*=true*/)
         else
             defaultImmSize += size/2;
     }
+    for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
+    {
+        CodeSpace *space = *i;
+        POLYUNSIGNED size = space->spaceSize();
+        defaultCodeSize += size/2;
+    }
     if (isExport)
     {
         // Minimum 1M words.
         if (defaultMutSize < 1024*1024) defaultMutSize = 1024*1024;
         if (defaultImmSize < 1024*1024) defaultImmSize = 1024*1024;
+        if (defaultCodeSize < 1024*1024) defaultCodeSize = 1024*1024;
     }
     else
     {
         // Much smaller minimum sizes for saved states.
         if (defaultMutSize < 1024) defaultMutSize = 1024;
         if (defaultImmSize < 4096) defaultImmSize = 4096;
+        if (defaultCodeSize < 4096) defaultImmSize = 4096;
     }
 }
 
@@ -191,11 +211,8 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
         return 0;
     // Ignore pointers to the IO area.  They will be relocated
     // when we write out the memory
-    MemSpace *space = gMem.SpaceForAddress(val.AsAddress());
+    MemSpace *space = gMem.SpaceForAddress(val.AsStackAddr()-1);
     ASSERT(space != 0);
-    if (space->spaceType == ST_IO)
-        return 0;
-
     // We may sometimes get addresses that have already been updated
     // to point to the new area.  e.g. (only?) in the case of constants
     // that have been updated in ScanConstantsWithinCode.
@@ -211,17 +228,6 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
             return 0;
     }
 
-    if (val.IsCodePtr())
-    {
-        // Find the start of the code segment
-        PolyObject *oldObject = ObjCodePtrToPtr(val.AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = val.AsCodePtr() - (byte*)oldObject;
-        PolyObject *newObject = ScanObjectAddress(oldObject);
-        *pt = PolyWord::FromCodePtr((byte*)newObject + offset);
-        return 0;
-    }
-
     ASSERT(OBJ_IS_DATAPTR(val));
 
     // Have we already scanned this?
@@ -254,7 +260,8 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     }
 
     // No, we need to copy it.
-    ASSERT(space->spaceType == ST_LOCAL || space->spaceType == ST_PERMANENT);
+    ASSERT(space->spaceType == ST_LOCAL || space->spaceType == ST_PERMANENT ||
+           space->spaceType == ST_CODE);
     POLYUNSIGNED lengthWord = obj->LengthWord();
     POLYUNSIGNED words = OBJ_OBJECT_LENGTH(lengthWord);
 
@@ -262,18 +269,21 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     bool isMutableObj = obj->IsMutable();
     bool isNoOverwrite = false;
     bool isByteObj = false;
+    bool isCodeObj = false;
     if (isMutableObj)
     {
         isNoOverwrite = obj->IsNoOverwriteObject();
         isByteObj = obj->IsByteObject();
     }
+    else isCodeObj = obj->IsCodeObject();
     // Allocate a new address for the object.
-    for (unsigned i = 0; i < gMem.neSpaces; i++)
+    for (std::vector<PermanentMemSpace *>::iterator i = gMem.eSpaces.begin(); i < gMem.eSpaces.end(); i++)
     {
-        PermanentMemSpace *space = gMem.eSpaces[i];
+        PermanentMemSpace *space = *i;
         if (isMutableObj == space->isMutable &&
             isNoOverwrite == space->noOverwrite &&
-            isByteObj == space->byteOnly)
+            isByteObj == space->byteOnly &&
+            isCodeObj == space->isCode)
         {
             ASSERT(space->topPointer <= space->top && space->topPointer >= space->bottom);
             POLYUNSIGNED spaceLeft = space->top - space->topPointer;
@@ -289,12 +299,19 @@ POLYUNSIGNED CopyScan::ScanAddressAt(PolyWord *pt)
     {
         // Didn't find room in the existing spaces.  Create a new space.
         POLYUNSIGNED spaceWords;
-        if (!isMutableObj) spaceWords = defaultImmSize;
-        else if (isNoOverwrite) spaceWords = defaultNoOverSize;
-        else spaceWords = defaultMutSize;
+        if (isMutableObj)
+        {
+            if (isNoOverwrite) spaceWords = defaultNoOverSize;
+            else spaceWords = defaultMutSize;
+        }
+        else
+        {
+            if (isCodeObj) spaceWords = defaultCodeSize;
+            else spaceWords = defaultImmSize;
+        }
         if (spaceWords <= words)
             spaceWords = words+1; // Make sure there's space for this object.
-        PermanentMemSpace *space = gMem.NewExportSpace(spaceWords, isMutableObj, isNoOverwrite);
+        PermanentMemSpace *space = gMem.NewExportSpace(spaceWords, isMutableObj, isNoOverwrite, isCodeObj);
         if (isByteObj) space->byteOnly = true;
         if (space == 0)
         {
@@ -354,14 +371,6 @@ PolyObject *CopyScan::ScanObjectAddress(PolyObject *base)
     return val.AsObjPtr();
 }
 
-// We clear the data in weak byte ref cells when they are exported
-// to an object file and also when they are read from a saved state.
-void ClearWeakByteRef::ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord)
-{
-    if (OBJ_IS_MUTABLE_OBJECT(lengthWord) && OBJ_IS_BYTE_OBJECT(lengthWord) && OBJ_IS_WEAKREF_OBJECT(lengthWord))
-        memset(base, 0, base->Length() * sizeof(PolyWord));
-}
-
 #define MAX_EXTENSION   4 // The longest extension we may need to add is ".obj"
 
 // Convert the forwarding pointers in a region back into length words.
@@ -379,7 +388,7 @@ static POLYUNSIGNED GetObjLength(PolyObject *obj)
     {
         PolyObject *forwardedTo = obj->GetForwardingPtr();
         POLYUNSIGNED length = GetObjLength(forwardedTo);
-        MemSpace *space = gMem.SpaceForAddress(forwardedTo);
+        MemSpace *space = gMem.SpaceForAddress(forwardedTo-1);
         if (space->spaceType == ST_EXPORT)
             obj->SetLengthWord(length);
         return length;
@@ -414,16 +423,16 @@ public:
     Exporter *exporter;
 };
 
-static void exporter(TaskData *taskData, Handle args, const TCHAR *extension, Exporter *exports)
+static void exporter(TaskData *taskData, Handle fileName, Handle root, const TCHAR *extension, Exporter *exports)
 {
-    TCHAR fileNameBuff[MAXPATHLEN+MAX_EXTENSION];
-    POLYUNSIGNED length =
-        Poly_string_to_C(DEREFHANDLE(args)->Get(0), fileNameBuff, MAXPATHLEN);
-    if (length > MAXPATHLEN)
-        raise_syscall(taskData, "File name too long", ENAMETOOLONG);
+    size_t extLen = _tcslen(extension);
+    TempString fileNameBuff(Poly_string_to_T_alloc(fileName->Word(), extLen));
+    if (fileNameBuff == NULL)
+        raise_syscall(taskData, "Insufficient memory", NOMEMORY);
+    size_t length = _tcslen(fileNameBuff);
 
     // Does it already have the extension?  If not add it on.
-    if (length < _tcslen(extension) || _tcscmp(fileNameBuff + length - _tcslen(extension), extension) != 0)
+    if (length < extLen || _tcscmp(fileNameBuff + length - extLen, extension) != 0)
         _tcscat(fileNameBuff, extension);
 #if (defined(_WIN32) && defined(UNICODE))
     exports->exportFile = _wfopen(fileNameBuff, L"wb");
@@ -431,9 +440,7 @@ static void exporter(TaskData *taskData, Handle args, const TCHAR *extension, Ex
     exports->exportFile = fopen(fileNameBuff, "wb");
 #endif
     if (exports->exportFile == NULL)
-        raise_syscall(taskData, "Cannot open export file", errno);
-
-    Handle root = taskData->saveVec.push(args->WordP()->Get(1));
+        raise_syscall(taskData, "Cannot open export file", ERRORNUMBER);
 
     // Request a full GC  to reduce the size of fix-ups.
     FullGC(taskData);
@@ -464,20 +471,25 @@ void Exporter::RunExport(PolyObject *rootFunction)
     }
 
     // Fix the forwarding pointers.
-    unsigned j;
-    for (j = 0; j < gMem.nlSpaces; j++)
+    for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *space = gMem.lSpaces[j];
+        LocalMemSpace *space = *i;
         // Local areas only have objects from the allocation pointer to the top.
         FixForwarding(space->bottom, space->lowerAllocPtr - space->bottom);
         FixForwarding(space->upperAllocPtr, space->top - space->upperAllocPtr);
     }
-    for (j = 0; j < gMem.npSpaces; j++)
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
     {
-        MemSpace *space = gMem.pSpaces[j];
+        MemSpace *space = *i;
         // Permanent areas are filled with objects from the bottom.
         FixForwarding(space->bottom, space->top - space->bottom);
     }
+    for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
+    {
+        MemSpace *space = *i;
+        // Code areas are filled with objects from the bottom.
+        FixForwarding(space->bottom, space->top - space->bottom);
+    }
 
     // Reraise the exception after cleaning up the forwarding pointers.
     if (copiedRoot == 0)
@@ -487,68 +499,52 @@ void Exporter::RunExport(PolyObject *rootFunction)
     }
 
     // Copy the areas into the export object.
-    unsigned tableEntries = gMem.neSpaces, memEntry = 0;
-    if (hierarchy != 0) tableEntries += gMem.npSpaces;
-    tableEntries++; // One extra for the IO area
+    size_t tableEntries = gMem.eSpaces.size();
+    unsigned memEntry = 0;
+    if (hierarchy != 0) tableEntries += gMem.pSpaces.size();
     exports->memTable = new memoryTableEntry[tableEntries];
-    exports->ioMemEntry = 0;
-
-    // The IO vector.  Should we actually create a blank area?  This needs to be
-    // writable by the RTS but not normally by ML.
-    MemSpace *ioSpace = gMem.IoSpace();
-    exports->memTable[0].mtAddr = ioSpace->bottom;
-    exports->memTable[0].mtLength = (char*)ioSpace->top - (char*)ioSpace->bottom;
-    exports->memTable[0].mtFlags = MTF_WRITEABLE; // Needs to be written during initialisation.
-    exports->memTable[0].mtIndex = 0;
-    memEntry = 1;
 
     // If we're constructing a module we need to include the global spaces.
     if (hierarchy != 0)
     {
         // Permanent spaces from the executable.
-        for (unsigned i = 0; i < gMem.npSpaces; i++)
+        for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
         {
-            PermanentMemSpace *space = gMem.pSpaces[i];
+            PermanentMemSpace *space = *i;
             if (space->hierarchy < hierarchy)
             {
                 memoryTableEntry *entry = &exports->memTable[memEntry++];
                 entry->mtAddr = space->bottom;
                 entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
                 entry->mtIndex = space->index;
-                entry->mtFlags = space->isMutable ? MTF_WRITEABLE : 0; 
+                entry->mtFlags = 0;
+                if (space->isMutable) entry->mtFlags |= MTF_WRITEABLE;
+                if (space->isCode) entry->mtFlags |= MTF_EXECUTABLE;
             }
         }
         newAreas = memEntry;
     }
 
-    for (unsigned i = 0; i < gMem.neSpaces; i++)
+    for (std::vector<PermanentMemSpace *>::iterator i = gMem.eSpaces.begin(); i < gMem.eSpaces.end(); i++)
     {
         memoryTableEntry *entry = &exports->memTable[memEntry++];
-        PermanentMemSpace *space = gMem.eSpaces[i];
+        PermanentMemSpace *space = *i;
         entry->mtAddr = space->bottom;
         entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
         entry->mtIndex = hierarchy == 0 ? memEntry-1 : space->index;
+        entry->mtFlags = 0;
         if (space->isMutable)
         {
             entry->mtFlags = MTF_WRITEABLE;
             if (space->noOverwrite) entry->mtFlags |= MTF_NO_OVERWRITE;
         }
-        else
-            entry->mtFlags = MTF_EXECUTABLE;
+        if (space->isCode) entry->mtFlags |= MTF_EXECUTABLE;
         if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
-        if (space->isMutable && space->byteOnly)
-        {
-            ClearWeakByteRef cwbr;
-            cwbr.ScanAddressesInRegion(space->bottom, space->topPointer);
-        }
     }
 
     ASSERT(memEntry == tableEntries);
     exports->memTableEntries = memEntry;
-
-    exports->ioSpacing = IO_SPACING;
     exports->rootFunction = copiedRoot;
-
     exports->exportStore();
     return;
 }
@@ -564,17 +560,20 @@ Handle exportNative(TaskData *taskData, Handle args)
     const char *extension = ".o"; // Cygwin
 #endif
     PECOFFExport exports;
-    exporter(taskData, args, extension, &exports);
+    exporter(taskData, taskData->saveVec.push(args->WordP()->Get(0)),
+        taskData->saveVec.push(args->WordP()->Get(1)), extension, &exports);
 #elif defined(HAVE_ELF_H) || defined(HAVE_ELF_ABI_H)
     // Most Unix including Linux, FreeBSD and Solaris.
     const char *extension = ".o";
     ELFExport exports;
-    exporter(taskData, args, extension, &exports);
+    exporter(taskData, taskData->saveVec.push(args->WordP()->Get(0)),
+        taskData->saveVec.push(args->WordP()->Get(1)), extension, &exports);
 #elif defined(HAVE_MACH_O_RELOC_H)
     // Mac OS-X
     const char *extension = ".o";
     MachoExport exports;
-    exporter(taskData, args, extension, &exports);
+    exporter(taskData, taskData->saveVec.push(args->WordP()->Get(0)),
+        taskData->saveVec.push(args->WordP()->Get(1)), extension, &exports);
 #else
     raise_exception_string (taskData, EXC_Fail, "Native export not available for this platform");
 #endif
@@ -584,10 +583,69 @@ Handle exportNative(TaskData *taskData, Handle args)
 Handle exportPortable(TaskData *taskData, Handle args)
 {
     PExport exports;
-    exporter(taskData, args, _T(".txt"), &exports);
+    exporter(taskData, taskData->saveVec.push(args->WordP()->Get(0)),
+        taskData->saveVec.push(args->WordP()->Get(1)), _T(".txt"), &exports);
     return taskData->saveVec.push(TAGGED(0));
 }
 
+POLYUNSIGNED PolyExport(PolyObject *threadId, PolyWord fileName, PolyWord root)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedName = taskData->saveVec.push(fileName);
+    Handle pushedRoot = taskData->saveVec.push(root);
+
+    try {
+#ifdef HAVE_PECOFF
+        // Windows including Cygwin
+#if (defined(_WIN32) && ! defined(__CYGWIN__))
+        const TCHAR *extension = _T(".obj"); // Windows
+#else
+        const char *extension = ".o"; // Cygwin
+#endif
+        PECOFFExport exports;
+        exporter(taskData, pushedName, pushedRoot, extension, &exports);
+#elif defined(HAVE_ELF_H) || defined(HAVE_ELF_ABI_H)
+        // Most Unix including Linux, FreeBSD and Solaris.
+        const char *extension = ".o";
+        ELFExport exports;
+        exporter(taskData, pushedName, pushedRoot, extension, &exports);
+#elif defined(HAVE_MACH_O_RELOC_H)
+        // Mac OS-X
+        const char *extension = ".o";
+        MachoExport exports;
+        exporter(taskData, pushedName, pushedRoot, extension, &exports);
+#else
+        raise_exception_string (taskData, EXC_Fail, "Native export not available for this platform");
+#endif
+    } catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    return TAGGED(0).AsUnsigned(); // Returns unit
+}
+
+POLYUNSIGNED PolyExportPortable(PolyObject *threadId, PolyWord fileName, PolyWord root)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedName = taskData->saveVec.push(fileName);
+    Handle pushedRoot = taskData->saveVec.push(root);
+
+    try {
+        PExport exports;
+        exporter(taskData, pushedName, pushedRoot, _T(".txt"), &exports);
+    } catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    return TAGGED(0).AsUnsigned(); // Returns unit
+}
+
 
 // Helper functions for exporting.  We need to produce relocation information
 // and this code is common to every method.
@@ -627,6 +685,16 @@ void Exporter::relocateObject(PolyObject *p)
 {
     if (p->IsByteObject())
     {
+        if (p->IsMutable() && p->IsWeakRefObject())
+        {
+            // Weak mutable byte refs are used for external references and
+            // also in the FFI for non-persistent values.
+            const char *entryName = getEntryPointName(p);
+            if (entryName != 0) addExternalReference(p, entryName);
+            // Clear the first word of the data.
+            ASSERT(p->Length() > 0);
+            p->Set(0, PolyWord::FromSigned(0));
+        }
     }
     else if (p->IsCodeObject())
     {
@@ -673,3 +741,10 @@ unsigned long ExportStringTable::makeEntry(const char *str)
     return entry;
 }
 
+struct _entrypts exporterEPT[] =
+{
+    { "PolyExport",                     (polyRTSFunction)&PolyExport},
+    { "PolyExportPortable",             (polyRTSFunction)&PolyExportPortable},
+
+    { NULL, NULL} // End of list.
+};
