@@ -30,10 +30,6 @@
 #include <stdio.h>
 #endif
 
-#ifdef HAVE_TIME_H
-#include <time.h>
-#endif
-
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
 #endif
@@ -57,6 +53,7 @@
 #include "processes.h" // For IO_SPACING
 #include "memmgr.h"
 #include "osmem.h"
+#include "rtsentry.h"
 
 /*
 This file contains the code both to export the file and to import it
@@ -103,41 +100,16 @@ unsigned long PExport::getIndex(PolyObject *p)
     }
 }
 
-void PExport::printCodeAddr(byte *q)
-// Address into code.  Either the pc field of a stack segment or
-//  a word + 2 format address.
-{
-    PolyObject *obj = ObjCodePtrToPtr(q);
-    unsigned long a = getIndex(obj);
-    fprintf(exportFile, "$%lu+%" POLYUFMT, a, (POLYUNSIGNED)(q - (byte*)obj));
-}
-
 /* Get the index corresponding to an address. */
 void PExport::printAddress(void *p)
 {
-    unsigned area = findArea(p);
-    if (area == ioMemEntry)
-    {
-        // Is it an IO entry?
-        POLYUNSIGNED byteOffset = (char*)p - (char*)memTable[area].mtAddr;
-        unsigned ioEntry = (unsigned)(byteOffset / (ioSpacing*sizeof(PolyWord)));
-        unsigned ioOffset = (unsigned)(byteOffset - ioEntry * (ioSpacing*sizeof(PolyWord)));
-        ASSERT(ioEntry < POLY_SYS_vecsize);
-        if (ioOffset == 0)
-           fprintf(exportFile, "I%d", ioEntry);
-        else
-            fprintf(exportFile, "J%d+%d", ioEntry, ioOffset); // Can this happen now?
-    }
-    else
-        fprintf(exportFile, "@%lu", getIndex((PolyObject*)p));
+    fprintf(exportFile, "@%lu", getIndex((PolyObject*)p));
 }
 
 void PExport::printValue(PolyWord q)
 {
     if (IS_INT(q) || q == PolyWord::FromUnsigned(0))
         fprintf(exportFile, "%" POLYSFMT, UNTAGGED(q));
-    else if (OBJ_IS_CODEPTR(q))
-        printCodeAddr(q.AsCodePtr());
     else
         printAddress(q.AsAddress());
 }
@@ -162,12 +134,15 @@ void PExport::printObject(PolyObject *p)
 
     if (p->IsByteObject())
     {
+        if (p->IsMutable() && p->IsWeakRefObject())
+        {
+            // This is either an entry point or a weak ref used in the FFI.
+            // Clear the first word
+            if (p->Length() >= 1) p->Set(0, PolyWord::FromSigned(0));
+        }
         /* May be a string, a long format arbitrary precision
            number or a real number. */
         PolyStringObject* ps = (PolyStringObject*)p;
-        /* See if the first word is a possible length.  The length
-           cannot be one because single character strings are
-           represented by the character. */
         /* This is not infallible but it seems to be good enough
            to detect the strings. */
         POLYUNSIGNED bytes = length * sizeof(PolyWord);
@@ -180,7 +155,7 @@ void PExport::printObject(PolyObject *p)
             for (unsigned i = 0; i < ps->length; i++)
             {
                 char ch = ps->chars[i];
-                fprintf(exportFile, "%02x", ch);
+                fprintf(exportFile, "%02x", ch & 0xff);
             }
         }
         else
@@ -258,13 +233,14 @@ void PExport::printObject(PolyObject *p)
 /* This is called for each constant within the code. 
    Print a relocation entry for the word and return a value that means
    that the offset is saved in original word. */
-void PExport::ScanConstant(byte *addr, ScanRelocationKind code)
+void PExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
     PolyWord p = GetConstantValue(addr, code);
     // We put in all the values including tagged constants.
-    PolyObject *obj = ObjCodePtrToPtr(addr);
     // Put in the byte offset and the relocation type code.
-    fprintf(exportFile, "%" POLYUFMT ",%d,", (POLYUNSIGNED)(addr - (byte*)obj), code);
+    POLYUNSIGNED offset = (POLYUNSIGNED)(addr - (byte*)base);
+    ASSERT (offset < base->Length() * sizeof(POLYUNSIGNED));
+    fprintf(exportFile, "%" POLYUFMT ",%d,", (POLYUNSIGNED)(addr - (byte*)base), code);
     printValue(p); // The value to plug in.
     fprintf(exportFile, " ");
 }
@@ -272,21 +248,16 @@ void PExport::ScanConstant(byte *addr, ScanRelocationKind code)
 void PExport::exportStore(void)
 {
     unsigned i;
-    time_t now;
-    time(&now);
 
     // Calculate a first guess for the map size based on the space size
     totalBytes = 0;
     void *startAddr = 0;
     for (i = 0; i < memTableEntries; i++)
     {
-        if (i != ioMemEntry)
-        {
-            totalBytes += (unsigned long)memTable[i].mtLength;
-            // Get the lowest address.
-            if (startAddr == 0 || memTable[i].mtAddr < startAddr)
-                startAddr = memTable[i].mtAddr;
-        }
+        totalBytes += (unsigned long)memTable[i].mtLength;
+        // Get the lowest address.
+        if (startAddr == 0 || memTable[i].mtAddr < startAddr)
+            startAddr = memTable[i].mtAddr;
     }
     // Create a map entry for each entry.  Allow five words per object.
     nMapSize = totalBytes/(sizeof(PolyWord)*5);
@@ -298,26 +269,23 @@ void PExport::exportStore(void)
     // We want the entries in pMap to be in ascending
     // order of address to make searching easy so we need to process the areas
     // in order of increasing address, which may not be the order in memTable.
-    indexOrder = (unsigned*)calloc(sizeof(unsigned), memTableEntries-1);
+    indexOrder = (unsigned*)calloc(sizeof(unsigned), memTableEntries);
     if (indexOrder == 0)
         throw MemoryException();
 
     unsigned items = 0;
     for (i = 0; i < memTableEntries; i++)
     {
-        if (i != ioMemEntry)
+        unsigned j = items;
+        while (j > 0 && memTable[i].mtAddr < memTable[indexOrder[j-1]].mtAddr)
         {
-            unsigned j = items;
-            while (j > 0 && memTable[i].mtAddr < memTable[indexOrder[j-1]].mtAddr)
-            {
-                indexOrder[j] = indexOrder[j-1];
-                j--;
-            }
-            indexOrder[j] = i;
-            items++;
+            indexOrder[j] = indexOrder[j-1];
+            j--;
         }
+        indexOrder[j] = i;
+        items++;
     }
-    ASSERT(items == memTableEntries-1);
+    ASSERT(items == memTableEntries);
 
     // Process the area in order of ascending address.
     for (i = 0; i < items; i++)
@@ -337,7 +305,6 @@ void PExport::exportStore(void)
                 if (newMap == 0)
                     throw MemoryException();
                 pMap = newMap;
-
             }
             POLYUNSIGNED length = obj->Length();
             pMap[nObjects++] = obj;
@@ -349,22 +316,18 @@ void PExport::exportStore(void)
     fprintf(exportFile, "Objects\t%lu\n", nObjects);
     fprintf(exportFile, "Root\t%lu\n", getIndex(rootFunction));
 
-
-    // Generate each of the areas apart from the IO area.
+    // Generate each of the areas.
     for (i = 0; i < memTableEntries; i++)
     {
-        if (i != ioMemEntry) // Don't relocate the IO area
+        char *start = (char*)memTable[i].mtAddr;
+        char *end = start + memTable[i].mtLength;
+        for (PolyWord *p = (PolyWord*)start; p < (PolyWord*)end; )
         {
-            char *start = (char*)memTable[i].mtAddr;
-            char *end = start + memTable[i].mtLength;
-            for (PolyWord *p = (PolyWord*)start; p < (PolyWord*)end; )
-            {
-                p++;
-                PolyObject *obj = (PolyObject*)p;
-                POLYUNSIGNED length = obj->Length();
-                printObject(obj);
-                p += length;
-            }
+            p++;
+            PolyObject *obj = (PolyObject*)p;
+            POLYUNSIGNED length = obj->Length();
+            printObject(obj);
+            p += length;
         }
     }
 
@@ -531,17 +494,16 @@ bool PImport::GetValue(PolyWord *result)
         /* IO entry number. */
         POLYUNSIGNED j;
         fscanf(f, "%" POLYUFMT, &j);
-        ASSERT(j < POLY_SYS_vecsize);
-        *result = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
-    }
-    else if (ch == 'J')
-    {
-        /* IO entry number with offset. */
-        POLYUNSIGNED j, offset;
-        fscanf(f, "%" POLYUFMT "+%" POLYUFMT, &j, &offset);
-        ASSERT(j < POLY_SYS_vecsize);
-        PolyWord base = (PolyObject*)&gMem.ioSpace->bottom[j * IO_SPACING];
-        *result = PolyWord::FromCodePtr(base.AsCodePtr() + offset);
+        // We may still have references to the old empty string value (j == 48).
+        if (j == 48)
+        {
+            // This is a bit of a hack but it's only temporary.
+            PolyObject  *p = NewObject(1, false);
+            p->SetLengthWord(1, F_BYTE_OBJ);
+            p->Set(0, PolyWord::FromUnsigned(0));
+            *result = p;
+        }
+        else ASSERT(0);
     }
     else
     {
@@ -568,16 +530,8 @@ bool PImport::DoImport()
     int ch;
     POLYUNSIGNED objNo;
 
-    ASSERT(gMem.npSpaces == 0);
-    ASSERT(gMem.neSpaces == 0);
-    ASSERT(gMem.ioSpace->bottom == 0);
-    PolyWord *ioSpace = (PolyWord*)calloc(POLY_SYS_vecsize*IO_SPACING, sizeof(PolyWord));
-    if (ioSpace == 0)
-    {
-        fprintf(stderr, "Unable to allocate memory\n");
-        return false;
-    }
-    gMem.InitIOSpace(ioSpace, POLY_SYS_vecsize*IO_SPACING);
+    ASSERT(gMem.pSpaces.size() == 0);
+    ASSERT(gMem.eSpaces.size() == 0);
 
     ch = getc(f);
     /* Skip the "Mapping" line. */
@@ -745,6 +699,12 @@ bool PImport::DoImport()
                 }
                 ch = getc(f);
                 ASSERT(ch == '\n');
+                // If this is an entry point object set its value.
+                if (p->IsMutable() && p->IsWeakRefObject())
+                {
+                    bool loadEntryPt = setEntryPoint(p);
+                    ASSERT(loadEntryPt);
+                }
                 break;
             }
 
