@@ -1,12 +1,11 @@
 /*
     Title:      Process environment.
-    Copyright (c) 2000-8
+    Copyright (c) 2000-8, 2016
         David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -58,16 +57,25 @@
 #include <process.h>
 #endif
 
+#ifdef HAVE_ASSERT_H
+#include <assert.h>
+#define ASSERT(x) assert(x)
+#else
+#define ASSERT(x) 0
+#endif
+
 // Include this next before errors.h since in WinCE at least the winsock errors are defined there.
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
 #include <winsock2.h>
 #include <tchar.h>
+#define NOMEMORY        ERROR_NOT_ENOUGH_MEMORY
+#undef ENOMEM
 #else
 typedef char TCHAR;
 #define _tgetenv getenv
+#define NOMEMORY ENOMEM
 #endif
 
-
 #include "globals.h"
 #include "sys.h"
 #include "run_time.h"
@@ -84,8 +92,17 @@ typedef char TCHAR;
 #include "processes.h"
 #include "locking.h"
 #include "errors.h"
+#include "rtsentry.h"
+#include "version.h"
 
-#include "poly_specific.h" // For the functions that have been moved.
+extern "C" {
+    POLYEXTERNALSYMBOL void PolyFinish(PolyObject *threadId, PolyWord arg);
+    POLYEXTERNALSYMBOL void PolyTerminate(PolyObject *threadId, PolyWord arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyProcessEnvGeneral(PolyObject *threadId, PolyWord code, PolyWord arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyProcessEnvErrorName(PolyObject *threadId, PolyWord syserr);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyProcessEnvErrorMessage(PolyObject *threadId, PolyWord syserr);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyProcessEnvErrorFromString(PolyObject *threadId, PolyWord string);
+}
 
 #define SAVE(x) mdTaskData->saveVec.push(x)
 #define ALLOC(n) alloc_and_save(mdTaskData, n)
@@ -143,7 +160,7 @@ void CygwinSpawnRequest::Perform()
 
 #endif
 
-Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
+static Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 {
     unsigned c = get_C_unsigned(mdTaskData, DEREFWORDHANDLE(code));
     switch (c)
@@ -157,7 +174,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
     case 14: /* Return a string from the environment. */
         {
             TempString buff(args->Word());
-            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", ENOMEM);
+            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", NOMEMORY);
             TCHAR *res = _tgetenv(buff);
             if (res == NULL) raise_syscall(mdTaskData, "Not Found", 0);
             else return SAVE(C_string_to_Poly(mdTaskData, res));
@@ -172,15 +189,15 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
         }
 
     case 15: /* Return the success value. */
-        return Make_arbitrary_precision(mdTaskData, EXIT_SUCCESS);
+        return Make_fixed_precision(mdTaskData, EXIT_SUCCESS);
 
     case 16: /* Return a failure value. */
-        return Make_arbitrary_precision(mdTaskData, EXIT_FAILURE);
+        return Make_fixed_precision(mdTaskData, EXIT_FAILURE);
 
     case 17: /* Run command. */
         {
             TempString buff(args->Word());
-            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", ENOMEM);
+            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", NOMEMORY);
             int res = -1;
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
             // Windows.
@@ -225,7 +242,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 sigemptyset(&sigset);
                 sigprocmask(SIG_SETMASK, &sigset, 0);
                 // Reset other signals?
-                execve("/bin/sh", argv, environ);
+                execv("/bin/sh", argv);
                 _exit(1);
             }
 #endif
@@ -243,12 +260,12 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                             DWORD result;
                             BOOL fResult = GetExitCodeProcess((HANDLE)pid, &result);
                             if (! fResult)
-                                raise_syscall(mdTaskData, "Function system failed", -(int)GetLastError());
+                                raise_syscall(mdTaskData, "Function system failed", GetLastError());
                             CloseHandle((HANDLE)pid);
-                            return Make_arbitrary_precision(mdTaskData, result);
+                            return Make_fixed_precision(mdTaskData, result);
                         }
                     case WAIT_FAILED:
-                        raise_syscall(mdTaskData, "Function system failed", -(int)GetLastError());
+                        raise_syscall(mdTaskData, "Function system failed", GetLastError());
                     }
                     // Wait for the process to exit or for the timeout
                     WaitHandle waiter((HANDLE)pid);
@@ -279,7 +296,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                     throw;
                 }
             }
-            return Make_arbitrary_precision(mdTaskData, res);
+            return Make_fixed_precision(mdTaskData, res);
         }
 
     case 18: /* Register function to run at exit. */
@@ -292,7 +309,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 cell->Set(1, DEREFWORD(args));
                 at_exit_list = cell;
             }
-            return Make_arbitrary_precision(mdTaskData, 0);
+            return Make_fixed_precision(mdTaskData, 0);
         }
 
     case 19: /* Return the next function in the atExit list and set the
@@ -319,67 +336,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
 
         /************ Error codes **************/
 
-    case 2: /* Get the name of a numeric error message. */
-        {
-            char buff[40];
-            int e = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
-            Handle  res;
-            /* First look to see if we have the name in
-               the error table. They should generally all be
-               there. */
-            const char *errorMsg = stringFromErrorCode(e);
-            if (errorMsg != NULL)
-                return SAVE(C_string_to_Poly(mdTaskData, errorMsg));
-            /* We get here if there's an error which isn't in the table. */
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
-            /* In the Windows version we may have both errno values
-               and also GetLastError values.  We convert the latter into
-               negative values before returning them. */
-            if (e < 0)
-            {
-                sprintf(buff, "WINERROR%0d", -e);
-                res = SAVE(C_string_to_Poly(mdTaskData, buff));
-                return res;
-            }
-            else
-#endif
-            {
-                sprintf(buff, "ERROR%0d", e);
-                res = SAVE(C_string_to_Poly(mdTaskData, buff));
-            }
-            return res;
-        }
 
-    case 3: /* Get the explanatory message for an error. */
-        {
-            return errorMsg(mdTaskData, get_C_int(mdTaskData, DEREFWORDHANDLE(args)));
-        }
-
-    case 4: /* Try to convert an error string to an error number. */
-        {
-            char buff[40];
-            /* Get the string. */
-            Poly_string_to_C(DEREFWORD(args), buff, sizeof(buff));
-            /* Look the string up in the table. */
-            int err = 0;
-            if (errorCodeFromString(buff, &err))
-                return Make_arbitrary_precision(mdTaskData, err);
-            /* If we don't find it then it may have been a constructed
-               error name. */
-            if (strncmp(buff, "ERROR", 5) == 0)
-            {
-                int i = atoi(buff+5);
-                if (i > 0) return Make_arbitrary_precision(mdTaskData, i);
-            }
-#if (defined(_WIN32) && ! defined(__CYGWIN__))
-            if (strncmp(buff, "WINERROR", 8) == 0)
-            {
-                int i = atoi(buff+8);
-                if (i > 0) return Make_arbitrary_precision(mdTaskData, -i);
-            }
-#endif
-            return Make_arbitrary_precision(mdTaskData, 0);
-        }
 
         /************ Directory/file paths **************/
 
@@ -397,23 +354,23 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
         {
             int e = get_C_int(mdTaskData, DEREFWORDHANDLE(args));
             if (ISPATHSEPARATOR(e))
-                return Make_arbitrary_precision(mdTaskData, 1);
-            else return Make_arbitrary_precision(mdTaskData, 0);
+                return Make_fixed_precision(mdTaskData, 1);
+            else return Make_fixed_precision(mdTaskData, 0);
         }
 
     case 9: /* Are names case-sensitive? */
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
         /* Windows - no. */
-        return Make_arbitrary_precision(mdTaskData, 0);
+        return Make_fixed_precision(mdTaskData, 0);
 #else
         /* Unix - yes. */
-        return Make_arbitrary_precision(mdTaskData, 1);
+        return Make_fixed_precision(mdTaskData, 1);
 #endif
 
         // These are no longer used.  The code is handled entirely in ML.
     case 10: /* Are empty arcs redundant? */
         /* Unix and Windows - yes. */
-        return Make_arbitrary_precision(mdTaskData, 1);
+        return Make_fixed_precision(mdTaskData, 1);
 
     case 11: /* Match the volume name part of a path. */
         {
@@ -432,7 +389,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             */
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
             TempString buff(path);
-            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", ENOMEM);
+            if (buff == 0) raise_syscall(mdTaskData, "Insufficient memory", NOMEMORY);
             size_t length = _tcslen(buff);
             if (length >= 2 && buff[1] == ':')
             { /* Volume name? */
@@ -512,7 +469,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
             if (IS_INT(volName))
             {
                 if (volName == TAGGED(0))
-                    return Make_arbitrary_precision(mdTaskData, 0);
+                    return Make_fixed_precision(mdTaskData, 0);
             }
             else
             {
@@ -520,7 +477,7 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 for (POLYUNSIGNED i = 0; i < volume->length; i++)
                 {
                     if (volume->chars[i] == '\0')
-                        return Make_arbitrary_precision(mdTaskData, 0);
+                        return Make_fixed_precision(mdTaskData, 0);
                 }
             }
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
@@ -537,23 +494,47 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
                 // This currently breaks the build.
                 case '/':
 #endif
-                    return Make_arbitrary_precision(mdTaskData, 0);
+                    return Make_fixed_precision(mdTaskData, 0);
                 }
-                if (*p >= 0 && *p <= 31) return Make_arbitrary_precision(mdTaskData, 0);
+                if (*p >= 0 && *p <= 31) return Make_fixed_precision(mdTaskData, 0);
             }
             // Should we check for special names such as aux, con, prn ??
-            return Make_arbitrary_precision(mdTaskData, 1);
+            return Make_fixed_precision(mdTaskData, 1);
 #else
             // That's all we need for Unix.
             // TODO: Check for /.  It's invalid in a file name arc.
-            return Make_arbitrary_precision(mdTaskData, 1);
+            return Make_fixed_precision(mdTaskData, 1);
 #endif
         }
 
-        // A group of calls have now been moved to poly_specific.
-        // This entry is returned for backwards compatibility.
-    case 100: case 101: case 102: case 103: case 104: case 105:
-        return poly_dispatch_c(mdTaskData, args, code);
+        // These were supposed to have been moved to poly-specific but don't seem to have been.
+    case 100: /* Return the maximum word segment size. */
+            return mdTaskData->saveVec.push(TAGGED(MAX_OBJECT_SIZE));
+    case 101: /* Return the maximum string size (in bytes).
+                 It is the maximum number of bytes in a segment
+                 less one word for the length field. */
+            return mdTaskData->saveVec.push(TAGGED((MAX_OBJECT_SIZE)*sizeof(PolyWord) - sizeof(PolyWord)));
+    // This is no longer in use but still seems to be in the pre-built compilers.
+    case 102: /* Test whether the supplied address is in the io area. */
+            return Make_arbitrary_precision(mdTaskData, 0);
+ 
+    case 104: return Make_arbitrary_precision(mdTaskData, POLY_version_number);
+ 
+    case 105: /* Get the name of the function. */
+        {
+            PolyObject *pt = DEREFWORDHANDLE(args);
+            if (pt->IsCodeObject()) /* Should now be a code object. */ 
+            {
+                /* Compiled code.  This is the first constant in the constant area. */
+                PolyWord *codePt = pt->ConstPtrForCode();
+                PolyWord name = codePt[0];
+                /* May be zero indicating an anonymous segment - return null string. */
+                if (name == PolyWord::FromUnsigned(0))
+                    return SAVE(C_string_to_Poly(mdTaskData, ""));
+                else return SAVE(name);
+            }
+            else raise_syscall(mdTaskData, "Not a code pointer", 0);
+        }
 
     default:
         {
@@ -565,18 +546,148 @@ Handle process_env_dispatch_c(TaskData *mdTaskData, Handle args, Handle code)
     }
 }
 
-/* Terminate normally with a result code. */
-Handle finishc(TaskData *taskData, Handle h)
+// General interface to process-env.  Ideally the various cases will be made into
+// separate functions.
+POLYUNSIGNED PolyProcessEnvGeneral(PolyObject *threadId, PolyWord code, PolyWord arg)
 {
-    int i = get_C_int(taskData, DEREFWORDHANDLE(h));
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedCode = taskData->saveVec.push(code);
+    Handle pushedArg = taskData->saveVec.push(arg);
+    Handle result = 0;
+
+    try {
+        result = process_env_dispatch_c(taskData, pushedArg, pushedCode);
+    }
+    catch (KillException &) {
+        processes->ThreadExit(taskData); // May test for kill
+    }
+    catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+// Terminate normally with a result code.
+void PolyFinish(PolyObject *threadId, PolyWord arg)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    int i = get_C_int(taskData, arg);
     // Cause the other threads to exit.
     processes->Exit(i);
     // Exit this thread
     processes->ThreadExit(taskData); // Doesn't return.
-    // Push a dummy result to keep lint happy
-    return taskData->saveVec.push(TAGGED(0));
 }
 
+// Terminate without running the atExit list or flushing buffers
+void PolyTerminate(PolyObject *threadId, PolyWord arg)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    int i = get_C_int(taskData, arg);
+    _exit(i); // Doesn't return.
+}
+
+// Get the name of a numeric error message.
+POLYUNSIGNED PolyProcessEnvErrorName(PolyObject *threadId, PolyWord syserr)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle result = 0;
+
+    try {
+        int e = (int)syserr.AsObjPtr()->Get(0).AsSigned();
+        // First look to see if we have the name in the error table. They should generally all be there.
+        const char *errorMsg = stringFromErrorCode(e);
+        if (errorMsg != NULL)
+            result = taskData->saveVec.push(C_string_to_Poly(taskData, errorMsg));
+        else
+        { // If it isn't in the table.
+            char buff[40];
+            sprintf(buff, "ERROR%0d", e);
+            result = taskData->saveVec.push(C_string_to_Poly(taskData, buff));
+        }
+    }
+    catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+/* Get the explanatory message for an error. */
+POLYUNSIGNED PolyProcessEnvErrorMessage(PolyObject *threadId, PolyWord syserr)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle result = 0;
+
+    try {
+        result = errorMsg(taskData, (int)syserr.AsObjPtr()->Get(0).AsSigned());
+    }
+    catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+// Try to convert an error string to an error number.
+POLYUNSIGNED PolyProcessEnvErrorFromString(PolyObject *threadId, PolyWord string)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle result = 0;
+
+    try {
+        char buff[40];
+        // Get the string.
+        Poly_string_to_C(string, buff, sizeof(buff));
+        // Look the string up in the table.
+        int err = 0;
+        if (errorCodeFromString(buff, &err))
+            result = Make_sysword(taskData, err);
+        else if (strncmp(buff, "ERROR", 5) == 0)
+        // If we don't find it then it may have been a constructed error name.
+            result = Make_sysword(taskData, atoi(buff+5));
+        else result = Make_sysword(taskData, 0); // Return 0w0 if it isn't there.
+
+    }
+    catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset); // Ensure the save vec is reset
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+struct _entrypts processEnvEPT[] =
+{
+    { "PolyFinish",                     (polyRTSFunction)&PolyFinish},
+    { "PolyTerminate",                  (polyRTSFunction)&PolyTerminate},
+    { "PolyProcessEnvGeneral",          (polyRTSFunction)&PolyProcessEnvGeneral},
+    { "PolyProcessEnvErrorName",        (polyRTSFunction)&PolyProcessEnvErrorName},
+    { "PolyProcessEnvErrorMessage",     (polyRTSFunction)&PolyProcessEnvErrorMessage},
+    { "PolyProcessEnvErrorFromString",  (polyRTSFunction)&PolyProcessEnvErrorFromString},
+
+    { NULL, NULL} // End of list.
+};
+
 class ProcessEnvModule: public RtsModule
 {
 public:
