@@ -105,13 +105,12 @@ public:
     virtual ~TaskData();
 
     void FillUnusedSpace(void);
-    void GarbageCollect(ScanAddress *process);
-    virtual void GCStack(ScanAddress *process) = 0;
+    virtual void GarbageCollect(ScanAddress *process);
 
     virtual Handle EnterPolyCode() = 0; // Start running ML
 
     virtual void InterruptCode() = 0;
-    virtual bool GetPCandSPFromContext(SIGNALCONTEXT *context, PolyWord * &sp,  POLYCODEPTR &pc) = 0;
+    virtual bool AddTimeProfileCount(SIGNALCONTEXT *context) = 0;
     // Initialise the stack for a new thread.  The parent task object is passed in because any
     // allocation that needs to be made must be made in the parent.
     virtual void InitStackFrame(TaskData *parentTask, Handle proc, Handle arg) = 0;
@@ -120,12 +119,10 @@ public:
     // ML callback function.
     virtual Handle EnterCallbackFunction(Handle func, Handle args) = 0;
 
-    virtual int  GetIOFunctionRegisterMask(int ioCall) = 0;
-
-    // Increment or decrement the first word of the object pointed to by the
-    // mutex argument and return the new value.
+    // The scheduler needs versions of atomic increment and atomic reset that
+    // work in exactly the same way as the code-generated versions (if any).
+    // Atomic decrement isn't needed since it only ever releases a mutex.
     virtual Handle AtomicIncrement(Handle mutexp) = 0;
-    virtual Handle AtomicDecrement(Handle mutexp) = 0;
     // Reset a mutex to one.  This needs to be atomic with respect to the
     // atomic increment and decrement instructions.
     virtual void AtomicReset(Handle mutexp) = 0;
@@ -133,12 +130,14 @@ public:
     virtual void CopyStackFrame(StackObject *old_stack, POLYUNSIGNED old_length,
                                 StackObject *new_stack, POLYUNSIGNED new_length) = 0;
 
-    // Put these in for the moment.  A few functions in run_time.cpp rely on them.
-    virtual POLYCODEPTR pc(void) const = 0;
-    virtual PolyWord *sp(void) const = 0;
-    virtual PolyWord *hr(void) const = 0;
-    virtual void set_hr(PolyWord *hr) = 0;
+
     virtual POLYUNSIGNED currentStackSpace(void) const = 0;
+    // Add a count to the local function if we are using store profiling.
+    virtual void addProfileCount(POLYUNSIGNED words) = 0;
+
+    // Functions called before and after an RTS call.
+    virtual void PreRTSCall(void) {}
+    virtual void PostRTSCall(void) {}
 
     SaveVec     saveVec;
     PolyWord    *allocPointer;  // Allocation pointer - decremented towards...
@@ -149,10 +148,12 @@ public:
     ThreadObject *threadObject;  // Pointer to the thread object.
     int         lastError;      // Last error from foreign code.
     void        *signalStack;  // Stack to handle interrupts (Unix only)
-    bool        pendingInterrupt; // The thread should trap into the RTS soon.
     PolyWord    foreignStack;   // Stack of saved data used in call_sym_and_convert
     bool        inML;          // True when this is in ML, false in the RTS
 
+    // Get a TaskData pointer given the ML taskId.
+    static TaskData *FindTaskForId(PolyObject *taskId);
+
 private:
     // If a thread has to block it will block on this.
     PCondVar threadLock;
@@ -172,7 +173,7 @@ private:
 #ifdef HAVE_WINDOWS_H
     LONGLONG lastCPUTime; // Used for profiling
 #endif
-#ifdef HAVE_PTHREAD
+#if ((!defined(_WIN32) || defined(__CYGWIN__)) && defined(HAVE_LIBPTHREAD) && defined(HAVE_PTHREAD_H))
 public:
     bool threadExited;
 private:
@@ -188,12 +189,6 @@ private:
 
 NORETURNFN(extern Handle exitThread(TaskData *mdTaskData));
 
-Handle ProcessAtomicReset(TaskData *taskData, Handle mutexp);
-Handle ProcessAtomicIncrement(TaskData *taskData, Handle mutexp);
-Handle ProcessAtomicDecrement(TaskData *taskData, Handle mutexp);
-Handle ThreadSelf(TaskData *taskData);
-Handle ThreadDispatch(TaskData *taskData, Handle args, Handle code);
-
 class ScanAddress;
 
 // Indicate what the main thread is doing if the profile
@@ -353,6 +348,6 @@ extern unsigned NumberOfPhysicalProcessors(void);
 
 extern ProcessExternal *processes;
 
-#define IO_SPACING 8 // This is a bit of a mess.
+extern struct _entrypts processesEPT[];
 
 #endif
