@@ -1,12 +1,11 @@
 /*
     Title:      Quick copying garbage collector
 
-    Copyright (c) 2011-12 David C. J. Matthews
+    Copyright (c) 2011-12, 2016 David C. J. Matthews
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -107,6 +106,18 @@ private:
     unsigned nOwnedSpaces;
 };
 
+// This is used when scanning code areas.  If there are no mutable cells left we can clear
+// the mutable bit and we don't have to scan it again.
+class CodeCheck: public ScanAddress
+{
+public:
+    CodeCheck(): foundMutable(false) {}
+    virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }
+    virtual void ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord)
+        { if (OBJ_IS_MUTABLE_OBJECT(lengthWord)) foundMutable = true;  }
+    bool foundMutable;
+};
+
 // This uses the conditional exchange instruction to check and update
 // the forwarding pointer.  It uses a lock prefix so that if another
 // thread has updated it in the meantime it will not set it.
@@ -137,7 +148,7 @@ static bool atomiclySetForwarding(LocalMemSpace *space, POLYUNSIGNED *pt,
     POLYUNSIGNED result = InterlockedCompareExchange(address, update, testVal);
     return result == testVal;
 # endif
-#elif(defined(HOSTARCHITECTURE_X86) && defined(__GNUC__))
+#elif((defined(HOSTARCHITECTURE_X86) || defined(HOSTARCHITECTURE_X32)) && defined(__GNUC__))
     POLYUNSIGNED result;
     __asm__ __volatile__ (
         "lock; cmpxchgl %1,%2"
@@ -232,9 +243,9 @@ LocalMemSpace *RootScanner::FindSpace(POLYUNSIGNED n, bool isMutable)
     }
 
     // Find the space with the largest free area.
-    for (unsigned i = 0; i < gMem.nlSpaces; i++)
+    for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *sp = gMem.lSpaces[i];
+        LocalMemSpace *sp = *i;
         if (sp->isMutable == isMutable && !sp->allocationSpace &&
                 (lSpace == 0 || sp->freeSpace() > lSpace->freeSpace()))
             lSpace = sp;
@@ -283,9 +294,9 @@ LocalMemSpace *ThreadScanner::FindSpace(POLYUNSIGNED n, bool isMutable)
     if (taskID != 0)
     {
         // See if we can take a space that is currently unused.
-        for (unsigned i = 0; i < gMem.nlSpaces; i++)
+        for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
         {
-            lSpace = gMem.lSpaces[i];
+            lSpace = *i;
             if (lSpace->spaceOwner == 0 && lSpace->isMutable == isMutable &&
                 ! lSpace->allocationSpace && lSpace->freeSpace() > n /* At least n+1*/)
             {
@@ -315,7 +326,7 @@ POLYUNSIGNED QuickGCScanner::ScanAddressAt(PolyWord *pt)
         PolyWord val = *(--pt);
         if (! val.IsTagged())
         {
-            LocalMemSpace *space = gMem.LocalSpaceForAddress(val.AsAddress());
+            LocalMemSpace *space = gMem.LocalSpaceForAddress(val.AsStackAddr()-1);
 
             // We only copy it if it is in a local allocation space and not in the
             // "overflow" area of data that could not copied by the last full GC.
@@ -502,9 +513,9 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
 
     POLYUNSIGNED spaceBeforeGC = 0;
 
-    for(unsigned k = 0; k < gMem.nlSpaces; k++)
+    for(std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *lSpace = gMem.lSpaces[k];
+        LocalMemSpace *lSpace = *i;
         ASSERT (lSpace->top >= lSpace->upperAllocPtr);
         ASSERT (lSpace->upperAllocPtr >= lSpace->lowerAllocPtr);
         ASSERT (lSpace->lowerAllocPtr >= lSpace->bottom);
@@ -532,12 +543,28 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
     RootScanner rootScan;
     // Scan the permanent mutable areas.  This could be parallelised but it doesn't
     // appear to be worthwhile at the moment.
-    for (unsigned j = 0; j < gMem.npSpaces; j++)
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
     {
-        PermanentMemSpace *space = gMem.pSpaces[j];
+        PermanentMemSpace *space = *i;
         if (space->isMutable && ! space->byteOnly)
             rootScan.ScanAddressesInRegion(space->bottom, space->top);
     }
+    // Scan code spaces.  
+    for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
+    {
+        CodeSpace *space = *i;
+        // Spaces are mutable if any object has been added to the area since the last GC.
+        if (space->isMutable)
+        {
+            rootScan.ScanAddressesInRegion(space->bottom, space->top);
+            // Check to see if any of the objects are still mutable.  If they are
+            // we are still building the code and must rescan it on the next GC.
+            // If there aren't we don't need to unless another code object is added.
+            CodeCheck codeCheck;
+            codeCheck.ScanAddressesInRegion(space->bottom, space->top);
+            space->isMutable = codeCheck.foundMutable;
+        }
+    }
 
     // Scan RTS addresses.  This will include the thread stacks.
     GCModules(&rootScan);
@@ -559,9 +586,9 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
     // tasks while we are still adding tasks.  It is important that the values of
     // partialGCRootBase, partialGCRootTop and partialGCTop are properly initialised
     // for these new spaces.
-    for (unsigned l = 0; l < gMem.nlSpaces; l++)
+    for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *space = gMem.lSpaces[l];
+        LocalMemSpace *space = *i;
         space->partialGCRootTop = space->lowerAllocPtr; // Top of the roots
         space->partialGCScan = space->lowerAllocPtr; // Start of scanning for new data.
     }
@@ -578,7 +605,7 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
                 // allocate a new space and realloc gMem.lSpaces.  We have to drop
                 // the lock before calling AddWorkOrRunNow in case we "run now".
                 PLocker lock(&localTableLock);
-                if (l >= gMem.nlSpaces)
+                if (l >= gMem.lSpaces.size())
                     break;
                 space = gMem.lSpaces[l++];
             }
@@ -599,9 +626,9 @@ bool RunQuickGC(const POLYUNSIGNED wordsRequiredToAllocate)
         globalStats.setSize(PSS_ALLOCATION, 0);
         globalStats.setSize(PSS_ALLOCATION_FREE, 0);
         // If it succeeded the allocation areas are now empty.
-        for(unsigned l = 0; l < gMem.nlSpaces; l++)
+        for(std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
         {
-            LocalMemSpace *lSpace = gMem.lSpaces[l];
+            LocalMemSpace *lSpace = *i;
             POLYUNSIGNED free;
             if (lSpace->allocationSpace)
             {
