@@ -5,12 +5,11 @@
     Copyright (c) 2000
         Cambridge University Technical Services Limited
 
-    Further work copyright David C.J. Matthews 2011
+    Further work copyright David C.J. Matthews 2011, 2016
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    License version 2.1 as published by the Free Software Foundation.
     
     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -78,6 +77,13 @@
 #include <stdint.h>
 #endif
 
+#ifdef HAVE_ASSERT_H 
+#include <assert.h>
+#define ASSERT(x)   assert(x)
+#else
+#define ASSERT(x)
+#endif
+
 #include "globals.h"
 #include "run_time.h"
 #include "reals.h"
@@ -89,6 +95,7 @@
 #include "rts_module.h"
 #include "machine_dep.h"
 #include "processes.h"
+#include "rtsentry.h"
 
 /*
     The Standard Basis Library assumes IEEE representation for reals.  Among other
@@ -104,6 +111,38 @@
     DCJM: March 2000.
 */
 
+extern "C" {
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyRealBoxedToString(PolyObject *threadId, PolyWord arg, PolyWord mode, PolyWord digits);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyRealGeneral(PolyObject *threadId, PolyWord code, PolyWord arg);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyRealBoxedFromString(PolyObject *threadId, PolyWord str);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyRealBoxedToLongInt(PolyObject *threadId, PolyWord arg);
+    POLYEXTERNALSYMBOL double PolyRealSqrt(double arg);
+    POLYEXTERNALSYMBOL double PolyRealSin(double arg);
+    POLYEXTERNALSYMBOL double PolyRealCos(double arg);
+    POLYEXTERNALSYMBOL double PolyRealArctan(double arg);
+    POLYEXTERNALSYMBOL double PolyRealExp(double arg);
+    POLYEXTERNALSYMBOL double PolyRealLog(double arg);
+    POLYEXTERNALSYMBOL double PolyRealTan(double arg);
+    POLYEXTERNALSYMBOL double PolyRealArcSin(double arg);
+    POLYEXTERNALSYMBOL double PolyRealArcCos(double arg);
+    POLYEXTERNALSYMBOL double PolyRealLog10(double arg);
+    POLYEXTERNALSYMBOL double PolyRealSinh(double arg);
+    POLYEXTERNALSYMBOL double PolyRealCosh(double arg);
+    POLYEXTERNALSYMBOL double PolyRealTanh(double arg);
+    POLYEXTERNALSYMBOL double PolyRealFloor(double arg);
+    POLYEXTERNALSYMBOL double PolyRealCeil(double arg);
+    POLYEXTERNALSYMBOL double PolyRealTrunc(double arg);
+    POLYEXTERNALSYMBOL double PolyRealRound(double arg);
+    POLYEXTERNALSYMBOL double PolyFloatArbitraryPrecision(PolyWord arg);
+    POLYEXTERNALSYMBOL POLYSIGNED PolyGetRoundingMode(PolyWord);
+    POLYEXTERNALSYMBOL POLYSIGNED PolySetRoundingMode(PolyWord);
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyRealSize(PolyWord);
+}
+
+static Handle Real_strc(TaskData *mdTaskData, Handle hDigits, Handle hMode, Handle arg);
+static Handle Real_convc(TaskData *mdTaskData, Handle str);
+
+
 // Positive and negative infinities and (positive) NaN.
 double posInf, negInf, notANumber;
 
@@ -144,155 +183,145 @@ Handle real_result(TaskData *mdTaskData, double x)
     return mdTaskData->saveVec.push(v);
 }
 
-/* CALL_IO2(Real_add, REF, REF, NOIND) */
-Handle Real_addc(TaskData *mdTaskData, Handle y, Handle x)
+POLYEXTERNALSYMBOL double PolyFloatArbitraryPrecision(PolyWord arg)
 {
-    return real_result(mdTaskData, real_arg(x)+real_arg(y));
+    return get_arbitrary_precision_as_real(arg);
 }
 
-/* CALL_IO2(Real_sub, REF, REF, NOIND) */
-Handle Real_subc(TaskData *mdTaskData, Handle y, Handle x)
+// Convert a boxed real to a long precision int.
+POLYUNSIGNED PolyRealBoxedToLongInt(PolyObject *threadId, PolyWord arg)
 {
-    return real_result(mdTaskData, real_arg(x)-real_arg(y));
-}
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedArg = taskData->saveVec.push(arg);
 
-/* CALL_IO2(Real_mul, REF, REF, NOIND) */
-Handle Real_mulc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    return real_result(mdTaskData, real_arg(x)*real_arg(y));
-}
-
-/* CALL_IO2(Real_div, REF, REF, NOIND) */
-Handle Real_divc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x);
-    double dy = real_arg(y);
-    return real_result(mdTaskData, dx/dy);
-}
-
-/* CALL_IO1(Real_abs, REF, NOIND) */
-// This handles +/- NaN correctly.
-// TODO: That assumes that notANumber is positive which may not be true.
-Handle Real_absc(TaskData *mdTaskData, Handle x)
-{
-    double dx = real_arg(x);
-    if (isnan(dx)) return real_result(mdTaskData, notANumber);
-    else return real_result(mdTaskData, fabs(real_arg(x)));
-}
-
-/* CALL_IO1(Real_neg, REF, NOIND) */
-Handle Real_negc(TaskData *mdTaskData, Handle x)
-{
-    return real_result(mdTaskData, -real_arg(x));
-}
-
-/* The old Real_comp function isn't right for IEEE arithmetic. These
-   functions were added to implement these correctly.
-   On Windows, at any rate, the comparison operations do not necessarily
-   return false on unordered arguments so we have to explicitly test for NaN.
- */
-Handle Real_geqc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(0));
-    return mdTaskData->saveVec.push(dx >= dy ? TAGGED(1) : TAGGED(0));
-}
-
-Handle Real_leqc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(0));
-    return mdTaskData->saveVec.push(dx <= dy ? TAGGED(1) : TAGGED(0));
-}
-
-Handle Real_gtrc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(0));
-    return mdTaskData->saveVec.push(dx > dy ? TAGGED(1) : TAGGED(0));
-}
-
-Handle Real_lssc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(0));
-    return mdTaskData->saveVec.push(dx < dy ? TAGGED(1) : TAGGED(0));
-}
-
-Handle Real_eqc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(0));
-    return mdTaskData->saveVec.push(dx == dy ? TAGGED(1) : TAGGED(0));
-}
-
-Handle Real_neqc(TaskData *mdTaskData, Handle y, Handle x)
-{
-    double dx = real_arg(x), dy = real_arg(y);
-    if (isnan(dx) || isnan(dy)) return mdTaskData->saveVec.push(TAGGED(1));
-    return mdTaskData->saveVec.push(dx != dy ? TAGGED(1) : TAGGED(0));
-}
-
-/* CALL_IO1(Real_float, REF, NOIND) */
-Handle Real_floatc(TaskData *mdTaskData, Handle x) /* SHORT int to real */
-{
-    double d = get_C_real(mdTaskData, DEREFWORDHANDLE(x));
-    return real_result(mdTaskData, d);
-}
-
-/* CALL_IO1(Real_int, REF, NOIND) */
-// This previously always converted to a SHORT integer.  Using
-// int64_t here means we will capture all the significant bits of
-// the mantissa.  The calling code checks for infinities and NaNs
-// and reduces the exponent if it is too big to fit.
-Handle Real_intc(TaskData *mdTaskData, Handle x)
-{
-    double dx = real_arg(x);
+    double dx = real_arg(pushedArg);
     int64_t i = (int64_t)dx;
-    return Make_arbitrary_precision(mdTaskData, i);
+    Handle result = Make_arbitrary_precision(taskData, i);
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
 }
 
-/* CALL_IO1(Real_sqrt, REF, NOIND) */
-Handle Real_sqrtc(TaskData *mdTaskData, Handle arg)
+// RTS call for square-root.
+double PolyRealSqrt(double arg)
 {
-    double dx = real_arg(arg);
-    return real_result(mdTaskData, sqrt(dx));
+    return sqrt(arg);
 }
 
-/* CALL_IO1(Real_sin, REF, NOIND) */
-Handle Real_sinc(TaskData *mdTaskData, Handle arg)
+// RTS call for sine.
+double PolyRealSin(double arg)
 {
-    return real_result(mdTaskData, sin(real_arg(arg)));
+    return sin(arg);
 }
 
-/* CALL_IO1(Real_cos, REF, NOIND) */
-Handle Real_cosc(TaskData *mdTaskData, Handle arg)
+// RTS call for cosine.
+double PolyRealCos(double arg)
 {
-    return real_result(mdTaskData, cos(real_arg(arg)));
+    return cos(arg);
 }
 
-/* CALL_IO1(Real_arctan, REF, NOIND) */
-Handle Real_arctanc(TaskData *mdTaskData, Handle arg)
+// RTS call for arctan.
+double PolyRealArctan(double arg)
 {
-    return real_result(mdTaskData, atan(real_arg(arg)));
+    return atan(arg);
 }
 
-/* CALL_IO1(Real_exp, REF, NOIND) */
-Handle Real_expc(TaskData *mdTaskData, Handle arg)
+// RTS call for exp.
+double PolyRealExp(double arg)
 {
-    return real_result(mdTaskData, exp(real_arg(arg)));
+    return exp(arg);
 }
 
-/* CALL_IO1(Real_ln, REF, NOIND) */
-Handle Real_lnc(TaskData *mdTaskData, Handle arg)
+// RTS call for ln.
+double PolyRealLog(double arg)
 {
-    double x = real_arg(arg);
-    /* Make sure the result conforms to the definition. */
-    if (x < 0.0)
-        return real_result(mdTaskData, notANumber); /* Nan. */
-    else if (x == 0.0) /* x may be +0.0 or -0.0 */
-        return real_result(mdTaskData, negInf); /* -infinity. */
-    else return real_result(mdTaskData, log(x));
+    // Make sure the result conforms to the definition.
+    // If the argument is a Nan each of the first two tests will fail.
+    if (arg > 0.0)
+        return log(arg);
+    else if (arg == 0.0) // x may be +0.0 or -0.0
+        return negInf; // -infinity.
+    else return notANumber;
+}
+
+// These were handled by the general dispatch function
+double PolyRealTan(double arg)
+{
+    return tan(arg);
+}
+
+double PolyRealArcSin(double arg)
+{
+    if (arg >= -1.0 && arg <= 1.0)
+        return asin(arg);
+    else return notANumber;
+}
+
+double PolyRealArcCos(double arg)
+{
+    if (arg >= -1.0 && arg <= 1.0)
+        return acos(arg);
+    else return notANumber;
+}
+
+double PolyRealLog10(double arg)
+{
+    // Make sure the result conforms to the definition.
+    // If the argument is a Nan each of the first two tests will fail.
+    if (arg > 0.0)
+        return log10(arg);
+    else if (arg == 0.0) // x may be +0.0 or -0.0
+        return negInf; // -infinity.
+    else return notANumber;
+}
+
+double PolyRealSinh(double arg)
+{
+    return sinh(arg);
+}
+
+double PolyRealCosh(double arg)
+{
+    return cosh(arg);
+}
+
+double PolyRealTanh(double arg)
+{
+    return tanh(arg);
+}
+
+double PolyRealFloor(double arg)
+{
+    return floor(arg);
+}
+
+double PolyRealCeil(double arg)
+{
+    return ceil(arg);
+}
+
+double PolyRealTrunc(double arg)
+{
+    // Truncate towards zero
+    if (arg >= 0.0) return floor(arg);
+    else return ceil(arg);
+}
+
+double PolyRealRound(double arg)
+{
+    // Round to nearest integral value.
+    double drem = fmod(arg, 2.0);
+    if (drem == 0.5 || drem == -1.5)
+        // If the value was exactly positive even + 0.5 or
+        // negative odd -0.5 round it down, otherwise round it up. 
+        return ceil(arg-0.5);
+    else return floor(arg+0.5);
+
 }
 
 /* CALL_IO1(Real_conv, REF, NOIND) */
@@ -301,7 +330,7 @@ Handle Real_convc(TaskData *mdTaskData, Handle str) /* string to real */
     double result;
     int i;
     char *finish;
-    char *string_buffer = Poly_string_to_C_alloc(DEREFHANDLE(str));
+    TempCString string_buffer(Poly_string_to_C_alloc(DEREFHANDLE(str)));
     
     /* Scan the string turning '~' into '-' */
     for(i = 0; string_buffer[i] != '\0'; i ++)
@@ -310,16 +339,38 @@ Handle Real_convc(TaskData *mdTaskData, Handle str) /* string to real */
     }
         
     /* Now convert it */
+#ifdef HAVE_STRTOD
+    result = strtod(string_buffer, &finish);
+#else
     result = poly_strtod(string_buffer, &finish);
-    bool isError = *finish != '\0'; // Test before deallocating
-    free(string_buffer);
+#endif
     // We no longer detect overflow and underflow and instead return
     // (signed) zeros for underflow and (signed) infinities for overflow.
-    if (isError) raise_exception_string(mdTaskData, EXC_conversion, "");
+    if (*finish != '\0') raise_exception_string(mdTaskData, EXC_conversion, "");
 
     return real_result(mdTaskData, result);
 }/* Real_conv */
 
+// Convert a string to a boxed real.  This should really return an unboxed real.
+POLYUNSIGNED PolyRealBoxedFromString(PolyObject *threadId, PolyWord str)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedString = taskData->saveVec.push(str);
+    Handle result = 0;
+
+    try {
+        result = Real_convc(taskData, pushedString);
+    } catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
 static double real_arg1(Handle x)
 {
     union db r_arg_x;
@@ -369,37 +420,62 @@ static Handle powerOf(TaskData *mdTaskData, Handle args)
 #define POLY_ROUND_DOWNWARD     1
 #define POLY_ROUND_UPWARD       2
 #define POLY_ROUND_TOZERO       3
+#define POLY_ROUND_ERROR        -1
+
+#if defined(__SOFTFP__)
+// soft-float lacks proper rounding mode support
+// While some systems will support fegetround/fesetround, it will have no
+// effect on the actual rounding performed, as the software implementation only
+// ever rounds to nearest.
+static int getrounding()
+{
+    return POLY_ROUND_TONEAREST;
+}
+
+static int setrounding(int rounding)
+{
+    switch (rounding)
+    {
+    case POLY_ROUND_TONEAREST: return 0; // The only mode supported
+    }
+    return -1; // Error - unsupported
+}
 
 // It would be nice to be able to use autoconf to test for these as functions
 // but they are frequently inlined 
-#if defined(HAVE_FENV_H)
+#elif defined(HAVE_FENV_H)
 // C99 version.  This is becoming the most common.
-static int getrounding(TaskData *)
+static int getrounding()
 {
     switch (fegetround())
     {
     case FE_TONEAREST: return POLY_ROUND_TONEAREST;
+#ifndef HOSTARCHITECTURE_SH
     case FE_DOWNWARD: return POLY_ROUND_DOWNWARD;
     case FE_UPWARD: return POLY_ROUND_UPWARD;
+#endif
     case FE_TOWARDZERO: return POLY_ROUND_TOZERO;
     }
-    return 0; // Keep the compiler happy
+    return POLY_ROUND_ERROR;
 }
 
-static void setrounding(TaskData *taskData, Handle args)
+static int setrounding(int rounding)
 {
-    switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
+    switch (rounding)
     {
-    case POLY_ROUND_TONEAREST: fesetround(FE_TONEAREST); break; // Choose nearest
-    case POLY_ROUND_DOWNWARD: fesetround(FE_DOWNWARD); break; // Towards negative infinity
-    case POLY_ROUND_UPWARD: fesetround(FE_UPWARD); break; // Towards positive infinity
-    case POLY_ROUND_TOZERO: fesetround(FE_TOWARDZERO); break; // Truncate towards zero
+    case POLY_ROUND_TONEAREST: fesetround(FE_TONEAREST); return 0; // Choose nearest
+#ifndef HOSTARCHITECTURE_SH
+    case POLY_ROUND_DOWNWARD: fesetround(FE_DOWNWARD); return 0; // Towards negative infinity
+    case POLY_ROUND_UPWARD: fesetround(FE_UPWARD); return 0; // Towards positive infinity
+#endif
+    case POLY_ROUND_TOZERO: fesetround(FE_TOWARDZERO); return 0; // Truncate towards zero
+    default: return -1;
     }
 }
 
 #elif (defined(HAVE_IEEEFP_H) && ! defined(__CYGWIN__))
 // Older FreeBSD.  Cygwin has the ieeefp.h header but not the functions!
-static int getrounding(TaskData *)
+static int getrounding()
 {
     switch (fpgetround())
     {
@@ -407,24 +483,25 @@ static int getrounding(TaskData *)
     case FP_RM: return POLY_ROUND_DOWNWARD;
     case FP_RP: return POLY_ROUND_UPWARD;
     case FP_RZ: return POLY_ROUND_TOZERO;
-    default: return 0; /* Shouldn't happen. */ 
+    default: return POLY_ROUND_ERROR; /* Shouldn't happen. */ 
     }
 }
 
-static void setrounding(TaskData *taskData, Handle args)
+static int setrounding(int rounding)
 {
-    switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
+    switch (rounding)
     {
     case POLY_ROUND_TONEAREST: fpsetround(FP_RN); break; /* Choose nearest */
     case POLY_ROUND_DOWNWARD: fpsetround(FP_RM); break; /* Towards negative infinity */
     case POLY_ROUND_UPWARD: fpsetround(FP_RP); break; /* Towards positive infinity */
     case POLY_ROUND_TOZERO: fpsetround(FP_RZ); break; /* Truncate towards zero */
     }
+    return 0
 }
 
 #elif defined(_WIN32)
 // Windows version
-static int getrounding(TaskData *)
+static int getrounding()
 {
     switch (_controlfp(0,0) & _MCW_RC)
     {
@@ -433,23 +510,24 @@ static int getrounding(TaskData *)
     case _RC_UP: return POLY_ROUND_UPWARD;
     case _RC_CHOP: return POLY_ROUND_TOZERO;
     }
-    return 0; // Keep the compiler happy
+    return POLY_ROUND_ERROR;
 }
 
-static void setrounding(TaskData *mdTaskData, Handle args)
+static int setrounding(int rounding)
 {
-    switch (get_C_long(mdTaskData, DEREFWORDHANDLE(args)))
+    switch (rounding)
     {
-    case POLY_ROUND_TONEAREST: _controlfp(_RC_NEAR, _MCW_RC); break; // Choose nearest
-    case POLY_ROUND_DOWNWARD: _controlfp(_RC_DOWN, _MCW_RC); break; // Towards negative infinity
-    case POLY_ROUND_UPWARD: _controlfp(_RC_UP, _MCW_RC); break; // Towards positive infinity
-    case POLY_ROUND_TOZERO: _controlfp(_RC_CHOP, _MCW_RC); break; // Truncate towards zero
+    case POLY_ROUND_TONEAREST: _controlfp(_RC_NEAR, _MCW_RC); return 0; // Choose nearest
+    case POLY_ROUND_DOWNWARD: _controlfp(_RC_DOWN, _MCW_RC); return 0; // Towards negative infinity
+    case POLY_ROUND_UPWARD: _controlfp(_RC_UP, _MCW_RC); return 0; // Towards positive infinity
+    case POLY_ROUND_TOZERO: _controlfp(_RC_CHOP, _MCW_RC); return 0; // Truncate towards zero
     }
+    return -1;
 }
 
 #elif defined(_FPU_GETCW) && defined(_FPU_SETCW)
 // Older Linux version
-static int getrounding(TaskData *)
+static int getrounding()
 {
     fpu_control_t ctrl;
     _FPU_GETCW(ctrl);
@@ -460,15 +538,15 @@ static int getrounding(TaskData *)
     case _FPU_RC_UP: return POLY_ROUND_UPWARD;
     case _FPU_RC_ZERO: return POLY_ROUND_TOZERO;
     }
-    return 0; /* Never reached but this avoids warning message. */
+    return POLY_ROUND_ERROR; /* Never reached but this avoids warning message. */
 }
 
-static void setrounding(TaskData *taskData, Handle args)
+static int setrounding(int rounding)
 {
     fpu_control_t ctrl;
     _FPU_GETCW(ctrl);
     ctrl &= ~_FPU_RC_ZERO; /* Mask off any existing rounding. */
-    switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
+    switch (rounding)
     {
     case POLY_ROUND_TONEAREST: ctrl |= _FPU_RC_NEAREST;
     case POLY_ROUND_DOWNWARD: ctrl |= _FPU_RC_DOWN;
@@ -476,54 +554,32 @@ static void setrounding(TaskData *taskData, Handle args)
     case POLY_ROUND_TOZERO: ctrl |= _FPU_RC_ZERO;
     }
     _FPU_SETCW(ctrl);
-}
-
-#elif (defined(HOSTARCHITECTURE_PPC) && defined(MACOSX))
-// Older versions of the Mac OS X didn't have a suitable function.
-
-static void getround(union db *res)
-{
-    __asm__ ("mffs f0");
-    __asm__ ("stfd f0,0(r3)");
-}
-
-static int getrounding(TaskData *)
-{
-    union db roundingRes;
-    getround(&roundingRes);
-    switch (roundingRes.wrd[1] & 3)
-    {
-    case 0: return POLY_ROUND_TONEAREST; /* Choose nearest */
-    case 1: return POLY_ROUND_TOZERO; /* Round towards zero */
-    case 2: return POLY_ROUND_UPWARD; /* Towards positive infinity */
-    case 3: return POLY_ROUND_DOWNWARD; /* Towards negative infinity */
-    }
-    return 0; /* Never reached but this avoids warning message. */
-}
-
-static void setrounding(TaskData *taskData, Handle args)
-{
-    switch (get_C_long(taskData, DEREFWORDHANDLE(args)))
-    {
-    case POLY_ROUND_TONEAREST: __asm__("mtfsfi 7,0"); break; /* Choose nearest */
-    case POLY_ROUND_DOWNWARD: __asm__("mtfsfi 7,3"); break; /* Towards negative infinity */
-    case POLY_ROUND_UPWARD: __asm__("mtfsfi 7,2"); break; /* Towards positive infinity */
-    case POLY_ROUND_TOZERO: __asm__("mtfsfi 7,1"); break; /* Truncate towards zero */
-    }
+    return 0;
 }
 #else
 // Give up.
-static int getrounding(TaskData *mdTaskData)
+static int getrounding()
 {
-    raise_exception_string(mdTaskData, EXC_Fail, "Unable to get floating point rounding control");
+    return POLY_ROUND_ERROR;
 }
 
-static void setrounding(TaskData *mdTaskData, Handle)
+static int setrounding()
 {
-    raise_exception_string(mdTaskData, EXC_Fail, "Unable to set floating point rounding control");
+    return -1;
 }
 #endif
 
+POLYSIGNED PolyGetRoundingMode(PolyWord)
+{
+    // Get the rounding and turn the result into a tagged integer.
+    return TAGGED(getrounding()).AsSigned();
+}
+
+POLYSIGNED PolySetRoundingMode(PolyWord arg)
+{
+    return TAGGED(setrounding((int)arg.UnTagged())).AsSigned();
+}
+
 Handle Real_strc(TaskData *mdTaskData, Handle hDigits, Handle hMode, Handle arg)
 {
     double  dx = real_arg(arg);
@@ -546,47 +602,36 @@ Handle Real_strc(TaskData *mdTaskData, Handle hDigits, Handle hMode, Handle arg)
     return mdTaskData->saveVec.push(result);
 }
 
+// Convert boxed real to string.  This should be changed to use an unboxed real argument.
+POLYUNSIGNED PolyRealBoxedToString(PolyObject *threadId, PolyWord arg, PolyWord mode, PolyWord digits)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedArg = taskData->saveVec.push(arg);
+    Handle pushedMode = taskData->saveVec.push(mode);
+    Handle pushedDigits = taskData->saveVec.push(digits);
+    Handle result = 0;
+
+    try {
+        result = Real_strc(taskData, pushedDigits, pushedMode, pushedArg);
+    } catch (...) { } // Can this raise an exception?
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
 /* Functions added for Standard Basis Library are all indirected through here. */
-Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
+static Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
 {
     unsigned c = get_C_unsigned(mdTaskData, DEREFWORDHANDLE(code));
     switch (c)
     {
-    case 0: /* tan */ return real_result(mdTaskData, tan(real_arg(args)));
-    case 1: /* asin */
-        {
-            double x = real_arg(args);
-            if (x < -1.0 || x > 1.0)
-                return real_result(mdTaskData, notANumber);
-            else return real_result(mdTaskData, asin(x));
-        }
-    case 2: /* acos */
-        {
-            double x = real_arg(args);
-            if (x < -1.0 || x > 1.0)
-                return real_result(mdTaskData, notANumber);
-            else return real_result(mdTaskData, acos(x));
-        }
     case 3: /* atan2 */ return real_result(mdTaskData, atan2(real_arg1(args), real_arg2(args)));
     case 4: /* pow */ return powerOf(mdTaskData, args);
-    case 5: /* log10 */
-        {
-            double x = real_arg(args);
-            /* Make sure the result conforms to the definition. */
-            if (x < 0.0)
-                return real_result(mdTaskData, notANumber); /* Nan. */
-            else if (x == 0.0) /* x may be +0.0 or -0.0 */
-                return real_result(mdTaskData, negInf); /* -infinity. */
-            else return real_result(mdTaskData, log10(x));
-        }
-    case 6: /* sinh */ return real_result(mdTaskData, sinh(real_arg(args)));
-    case 7: /* cosh */ return real_result(mdTaskData, cosh(real_arg(args)));
-    case 8: /* tanh */ return real_result(mdTaskData, tanh(real_arg(args)));
-    case 9: /* setroundingmode */
-        setrounding(mdTaskData, args);
-        return mdTaskData->saveVec.push(TAGGED(0)); /* Unit */
-    case 10: /* getroundingmode */
-        return mdTaskData->saveVec.push(TAGGED(getrounding(mdTaskData)));
     /* Floating point representation queries. */
 #ifdef _DBL_RADIX
     case 11: /* Value of radix */ return mdTaskData->saveVec.push(TAGGED(_DBL_RADIX));
@@ -600,36 +645,10 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
        number which can be represented is DBL_MIN*2**(-DBL_MANT_DIG) */
     case 14: /* Minimum normalised number. */
         return real_result(mdTaskData, DBL_MIN);
-#if (0)
-    case 15: /* Is finite */ /* No longer used - implemented in ML. */
-        return mdTaskData->saveVec.push(finite(real_arg(args)) ? TAGGED(1) : TAGGED(0));
-    case 16: /* Is Nan */ /* No longer used - implemented in ML. */
-        return mdTaskData->saveVec.push(isnan(real_arg(args)) ? TAGGED(1) : TAGGED(0));
-#endif
     case 17: /* Get sign bit.  There may be better ways to find this. */
         return mdTaskData->saveVec.push(copysign(1.0, real_arg(args)) < 0.0 ? TAGGED(1) : TAGGED(0));
     case 18: /* Copy sign. */
         return real_result(mdTaskData, copysign(real_arg1(args), real_arg2(args)));
-    case 19: /* Return largest integral value (as a real) <= x. */
-        return real_result(mdTaskData, floor(real_arg(args)));
-    case 20: /* Return smallest integral value (as a real) >= x  */
-        return real_result(mdTaskData, ceil(real_arg(args)));
-    case 21:
-        { /* Truncate towards zero */
-            double dx = real_arg(args);
-            if (dx >= 0.0) return real_result(mdTaskData, floor(dx));
-            else return real_result(mdTaskData, ceil(dx));
-        }
-    case 22: /* Round to nearest integral value. */
-        {
-            double dx = real_arg(args);
-            double drem = fmod(dx, 2.0);
-            if (drem == 0.5 || drem == -1.5)
-                /* If the value was exactly positive even + 0.5 or
-                   negative odd -0.5 round it down, otherwise round it up. */
-                return real_result(mdTaskData, ceil(dx-0.5));
-            else return real_result(mdTaskData, floor(dx+0.5));
-        }
     case 23: /* Compute ldexp */
         {
             int exp = get_C_int(mdTaskData, DEREFHANDLE(args)->Get(1));
@@ -646,39 +665,6 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
             (void)frexp(real_arg(args), &exp);
             return mdTaskData->saveVec.push(TAGGED(exp));
         }
-    case 26: /* Return the mantissa from a Nan as a real number. */
-        // I think this is no longer used.
-        {
-            union db r_arg_x, r_arg_y;
-            /* We want to simply replace the exponent by the exponent
-               value for 0.5<=x<1.
-               I think there may be a more portable way of doing this. */
-            r_arg_x.dble = posInf; /* Positive infinity. */
-            r_arg_y.dble = 0.5;
-            /* Use the infinity value as a mask, removing any bits set
-               and replace by the exponent from 0.5. */
-            byte *barg = DEREFBYTEHANDLE(args);
-            for(unsigned i = 0; i < DBLE; i++)
-            {
-                r_arg_x.bytes[i] = (barg[i] & ~r_arg_x.bytes[i]) | r_arg_y.bytes[i];
-            }
-            return real_result(mdTaskData, r_arg_x.dble);
-        }
-    case 27: /* Construct a Nan from a given mantissa. */
-        // I think this is no longer used.
-        {
-            union db r_arg;
-            r_arg.dble = posInf; /* Positive infinity. */
-            /* OR in the exponent. */
-            byte *barg = DEREFBYTEHANDLE(args);
-            for(unsigned i = 0; i < DBLE; i++)
-            {
-                r_arg.bytes[i] = r_arg.bytes[i] | barg[i];
-            }
-            return real_result(mdTaskData, r_arg.dble);
-        }
-    case 28: /* Return the number of bytes for a real.  */
-        return mdTaskData->saveVec.push(TAGGED(sizeof(double)));
 
     default:
         {
@@ -690,6 +676,63 @@ Handle Real_dispatchc(TaskData *mdTaskData, Handle args, Handle code)
     }
 }
 
+POLYUNSIGNED PolyRealSize(PolyWord)
+{
+    // Return the number of bytes for a real.  This is used in PackRealBig/Little.
+    return TAGGED(sizeof(double)).AsUnsigned();
+}
+
+POLYUNSIGNED PolyRealGeneral(PolyObject *threadId, PolyWord code, PolyWord arg)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+    Handle pushedCode = taskData->saveVec.push(code);
+    Handle pushedArg = taskData->saveVec.push(arg);
+    Handle result = 0;
+
+    try {
+        result = Real_dispatchc(taskData, pushedArg, pushedCode);
+    } catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    if (result == 0) return TAGGED(0).AsUnsigned();
+    else return result->Word().AsUnsigned();
+}
+
+struct _entrypts realsEPT[] =
+{
+    { "PolyRealBoxedToString",          (polyRTSFunction)&PolyRealBoxedToString},
+    { "PolyRealGeneral",                (polyRTSFunction)&PolyRealGeneral},
+    { "PolyRealBoxedFromString",        (polyRTSFunction)&PolyRealBoxedFromString},
+    { "PolyRealBoxedToLongInt",         (polyRTSFunction)&PolyRealBoxedToLongInt},
+    { "PolyRealSqrt",                   (polyRTSFunction)&PolyRealSqrt},
+    { "PolyRealSin",                    (polyRTSFunction)&PolyRealSin},
+    { "PolyRealCos",                    (polyRTSFunction)&PolyRealCos},
+    { "PolyRealArctan",                 (polyRTSFunction)&PolyRealArctan},
+    { "PolyRealExp",                    (polyRTSFunction)&PolyRealExp},
+    { "PolyRealLog",                    (polyRTSFunction)&PolyRealLog},
+    { "PolyRealTan",                    (polyRTSFunction)&PolyRealTan},
+    { "PolyRealArcSin",                 (polyRTSFunction)&PolyRealArcSin},
+    { "PolyRealArcCos",                 (polyRTSFunction)&PolyRealArcCos},
+    { "PolyRealLog10",                  (polyRTSFunction)&PolyRealLog10},
+    { "PolyRealSinh",                   (polyRTSFunction)&PolyRealSinh},
+    { "PolyRealCosh",                   (polyRTSFunction)&PolyRealCosh},
+    { "PolyRealTanh",                   (polyRTSFunction)&PolyRealTanh},
+    { "PolyRealFloor",                  (polyRTSFunction)&PolyRealFloor},
+    { "PolyRealCeil",                   (polyRTSFunction)&PolyRealCeil},
+    { "PolyRealTrunc",                  (polyRTSFunction)&PolyRealTrunc},
+    { "PolyRealRound",                  (polyRTSFunction)&PolyRealRound},
+    { "PolyFloatArbitraryPrecision",    (polyRTSFunction)&PolyFloatArbitraryPrecision},
+    { "PolyGetRoundingMode",            (polyRTSFunction)&PolyGetRoundingMode},
+    { "PolySetRoundingMode",            (polyRTSFunction)&PolySetRoundingMode},
+    { "PolyRealSize",                   (polyRTSFunction)&PolyRealSize},
+
+    { NULL, NULL} // End of list.
+};
+
 class RealArithmetic: public RtsModule
 {
 public:
