@@ -72,6 +72,8 @@
 
 #if (defined(_WIN32) && ! defined(__CYGWIN__))
 #include <tchar.h>
+#define ERRORNUMBER _doserrno
+#define NOMEMORY ERROR_NOT_ENOUGH_MEMORY
 #else
 typedef char TCHAR;
 #define _T(x) x
@@ -84,6 +86,8 @@ typedef char TCHAR;
 #ifndef lstrcmpi
 #define lstrcmpi strcasecmp
 #endif
+#define ERRORNUMBER errno
+#define NOMEMORY ENOMEM
 #endif
 
 
@@ -100,10 +104,9 @@ typedef char TCHAR;
 #include "machine_dep.h"
 #include "osmem.h"
 #include "gc.h" // For FullGC.
-
-#if(!defined(MAXPATHLEN) && defined(MAX_PATH))
-#define MAXPATHLEN MAX_PATH
-#endif
+#include "timing.h"
+#include "rtsentry.h"
+#include "check_objects.h"
 
 #ifdef _MSC_VER
 // Don't tell me about ISO C++ changes.
@@ -144,7 +147,7 @@ private:
  */
 
 #define SAVEDSTATESIGNATURE "POLYSAVE"
-#define SAVEDSTATEVERSION   1
+#define SAVEDSTATEVERSION   2
 
 // File header for a saved state file.  This appears as the first entry
 // in the file.
@@ -164,9 +167,7 @@ typedef struct _savedStateHeader
     size_t      stringTableSize;        // Size of string table
     unsigned    parentNameEntry;        // Position of parent name in string table (0 if top)
     time_t      timeStamp;            // The time stamp for this file.
-    uintptr_t   fileSignature;        // The signature for this file.
     time_t      parentTimeStamp;      // The time stamp for the parent.
-    uintptr_t   parentSignature;      // The signature for the parent.
 } SavedStateHeader;
 
 // Entry for segment table.  This describes the segments on the disc that
@@ -186,7 +187,8 @@ typedef struct _savedStateSegmentDescr
 #define SSF_WRITABLE    1               // The segment contains mutable data
 #define SSF_OVERWRITE   2               // The segment overwrites the data (mutable) in a parent.
 #define SSF_NOOVERWRITE 4               // The segment must not be further overwritten
-#define SSF_BYTES       8               // The segment containing only byte data
+#define SSF_BYTES       8               // The segment contains only byte data
+#define SSF_CODE        16              // The segment contains only code
 
 typedef struct _relocationEntry
 {
@@ -285,7 +287,7 @@ public:
 
 private:
     // ScanAddress overrides
-    virtual void ScanConstant(byte *addrOfConst, ScanRelocationKind code);
+    virtual void ScanConstant(PolyObject *base, byte *addrOfConst, ScanRelocationKind code);
     // At the moment we should only get calls to ScanConstant.
     virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }
 
@@ -326,7 +328,7 @@ PolyWord SaveStateExport::createRelocation(PolyWord p, void *relocAddr)
 /* This is called for each constant within the code. 
    Print a relocation entry for the word and return a value that means
    that the offset is saved in original word. */
-void SaveStateExport::ScanConstant(byte *addr, ScanRelocationKind code)
+void SaveStateExport::ScanConstant(PolyObject *base, byte *addr, ScanRelocationKind code)
 {
     PolyWord p = GetConstantValue(addr, code);
 
@@ -378,6 +380,9 @@ protected:
     virtual PolyObject *ScanObjectAddress(PolyObject *base)
         { return GetNewAddress(base).AsObjPtr(); }
     PolyWord GetNewAddress(PolyWord old);
+
+public:
+    void ScanCodeSpace(CodeSpace *space);
 };
 
 
@@ -391,21 +396,9 @@ POLYUNSIGNED SaveFixupAddress::ScanAddressAt(PolyWord *pt)
 // has moved, otherwise returns the original.
 PolyWord SaveFixupAddress::GetNewAddress(PolyWord old)
 {
-    if (old.IsTagged() || old == PolyWord::FromUnsigned(0) || gMem.IsIOPointer(old.AsAddress()))
+    if (old.IsTagged() || old == PolyWord::FromUnsigned(0))
         return old; //  Nothing to do.
 
-    // When we are updating addresses in the stack or in code segments we may have
-    // code pointers.
-    if (old.IsCodePtr())
-    {
-        // Find the start of the code segment
-        PolyObject *oldObject = ObjCodePtrToPtr(old.AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = old.AsCodePtr() - (byte*)oldObject;
-        PolyWord newObject = GetNewAddress(oldObject);
-        return PolyWord::FromCodePtr(newObject.AsCodePtr() + offset);
-    }
-
     ASSERT(old.IsDataPtr());
 
     PolyObject *obj = old.AsObjPtr();
@@ -421,6 +414,24 @@ PolyWord SaveFixupAddress::GetNewAddress(PolyWord old)
     return old;
 }
 
+// Fix up addresses in the code area.  Unlike ScanAddressesInRegion this updates
+// cells that have been moved.  We need to do that because we may still have
+// return addresses into those cells and we don't move return addresses.  We
+// do want the code to see updated constant addresses.
+void SaveFixupAddress::ScanCodeSpace(CodeSpace *space)
+{
+    for (PolyWord *pt = space->bottom; pt < space->top; )
+    {
+        pt++;
+        PolyObject *obj = (PolyObject*)pt;
+        PolyObject *dest = obj->FollowForwardingChain();
+        POLYUNSIGNED length = dest->Length();
+        if (length != 0)
+            ScanAddressesInObject(obj, dest->LengthWord());
+        pt += length;
+    }
+}
+
 // Called by the root thread to actually save the state and write the file.
 void SaveRequest::Perform()
 {
@@ -440,7 +451,7 @@ void SaveRequest::Perform()
     if (exports.exportFile == NULL)
     {
         errorMessage = "Cannot open save file";
-        errCode = errno;
+        errCode = ERRORNUMBER;
         return;
     }
 
@@ -450,9 +461,9 @@ void SaveRequest::Perform()
     copyScan.initialise(false);
     bool success = true;
     try {
-        for (unsigned i = 0; i < gMem.npSpaces; i++)
+        for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
         {
-            PermanentMemSpace *space = gMem.pSpaces[i];
+            PermanentMemSpace *space = *i;
             if (space->isMutable && ! space->noOverwrite && ! space->byteOnly)
                 copyScan.ScanAddressesInRegion(space->bottom, space->top);
         }
@@ -464,71 +475,87 @@ void SaveRequest::Perform()
 
     // Copy the areas into the export object.  Make sufficient space for
     // the largest possible number of entries.
-    exports.memTable = new memoryTableEntry[gMem.neSpaces+gMem.npSpaces+1];
-    exports.ioMemEntry = 0;
-    // The IO vector.
+    exports.memTable = new memoryTableEntry[gMem.eSpaces.size()+gMem.pSpaces.size()+1];
     unsigned memTableCount = 0;
-    MemSpace *ioSpace = gMem.IoSpace();
-    exports.memTable[0].mtAddr = ioSpace->bottom;
-    exports.memTable[0].mtLength = (char*)ioSpace->top - (char*)ioSpace->bottom;
-    exports.memTable[0].mtFlags = 0;
-    exports.memTable[0].mtIndex = 0;
-    memTableCount++;
 
     // Permanent spaces at higher level.  These have to have entries although
     // only the mutable entries will be written.
-    for (unsigned w = 0; w < gMem.npSpaces; w++)
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
     {
-        PermanentMemSpace *space = gMem.pSpaces[w];
+        PermanentMemSpace *space = *i;
         if (space->hierarchy < newHierarchy)
         {
             memoryTableEntry *entry = &exports.memTable[memTableCount++];
             entry->mtAddr = space->bottom;
             entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
             entry->mtIndex = space->index;
+            entry->mtFlags = 0;
             if (space->isMutable)
             {
-                entry->mtFlags = MTF_WRITEABLE;
+                entry->mtFlags |= MTF_WRITEABLE;
                 if (space->noOverwrite) entry->mtFlags |= MTF_NO_OVERWRITE;
                 if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
             }
-            else
-                entry->mtFlags = MTF_EXECUTABLE;
+            if (space->isCode)
+                entry->mtFlags |= MTF_EXECUTABLE;
         }
     }
     unsigned permanentEntries = memTableCount; // Remember where new entries start.
 
     // Newly created spaces.
-    for (unsigned i = 0; i < gMem.neSpaces; i++)
+    for (std::vector<PermanentMemSpace *>::iterator i = gMem.eSpaces.begin(); i < gMem.eSpaces.end(); i++)
     {
         memoryTableEntry *entry = &exports.memTable[memTableCount++];
-        PermanentMemSpace *space = gMem.eSpaces[i];
+        PermanentMemSpace *space = *i;
         entry->mtAddr = space->bottom;
         entry->mtLength = (space->topPointer-space->bottom)*sizeof(PolyWord);
         entry->mtIndex = space->index;
+        entry->mtFlags = 0;
         if (space->isMutable)
         {
-            entry->mtFlags = MTF_WRITEABLE;
+            entry->mtFlags |= MTF_WRITEABLE;
             if (space->noOverwrite) entry->mtFlags |= MTF_NO_OVERWRITE;
             if (space->byteOnly) entry->mtFlags |= MTF_BYTES;
         }
-        else
-            entry->mtFlags = MTF_EXECUTABLE;
+        if (space->isCode)
+            entry->mtFlags |= MTF_EXECUTABLE;
     }
 
     exports.memTableEntries = memTableCount;
-    exports.ioSpacing = IO_SPACING;
 
     // Update references to moved objects.
     SaveFixupAddress fixup;
-    for (unsigned l = 0; l < gMem.nlSpaces; l++)
+    for (std::vector<LocalMemSpace*>::iterator i = gMem.lSpaces.begin(); i < gMem.lSpaces.end(); i++)
     {
-        LocalMemSpace *space = gMem.lSpaces[l];
+        LocalMemSpace *space = *i;
         fixup.ScanAddressesInRegion(space->bottom, space->lowerAllocPtr);
         fixup.ScanAddressesInRegion(space->upperAllocPtr, space->top);
     }
+    for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
+        fixup.ScanCodeSpace(*i);
+
     GCModules(&fixup);
 
+    // Restore the length words in the code areas.
+    // Although we've updated any pointers to the start of the code we could have return addresses
+    // pointing to the original code.  GCModules updates the stack but doesn't update return addresses.
+    for (std::vector<CodeSpace *>::iterator i = gMem.cSpaces.begin(); i < gMem.cSpaces.end(); i++)
+    {
+        CodeSpace *space = *i;
+        for (PolyWord *pt = space->bottom; pt < space->top; )
+        {
+            pt++;
+            PolyObject *obj = (PolyObject*)pt;
+            if (obj->ContainsForwardingPtr())
+            {
+                PolyObject *forwardedTo = obj->FollowForwardingChain();
+                POLYUNSIGNED lengthWord = forwardedTo->LengthWord();
+                obj->SetLengthWord(lengthWord);
+            }
+            pt += obj->Length();
+        }
+    }
+
     // Update the global memory space table.  Old segments at the same level
     // or lower are removed.  The new segments become permanent.
     // Try to promote the spaces even if we've had a failure because export
@@ -537,7 +564,7 @@ void SaveRequest::Perform()
     if (! gMem.PromoteExportSpaces(newHierarchy) || ! success)
     {
         errorMessage = "Out of Memory";
-        errCode = ENOMEM;
+        errCode = NOMEMORY;
         return;
     }
     // Remove any deeper entries from the hierarchy table.
@@ -563,7 +590,7 @@ void SaveRequest::Perform()
         saveHeader.parentTimeStamp = hierarchyTable[newHierarchy-2]->timeStamp;
         saveHeader.parentNameEntry = sizeof(TCHAR); // Always the first entry.
     }
-    saveHeader.timeStamp = time(NULL);
+    saveHeader.timeStamp = getBuildTime();
     saveHeader.segmentDescrCount = exports.memTableEntries; // One segment for each space.
     // Write out the header.
     fwrite(&saveHeader, sizeof(saveHeader), 1, exports.exportFile);
@@ -590,6 +617,8 @@ void SaveRequest::Perform()
             if (entry->mtFlags & MTF_BYTES)
                 descrs[j].segmentFlags |= SSF_BYTES;
         }
+        if (entry->mtFlags & MTF_EXECUTABLE)
+            descrs[j].segmentFlags |= SSF_CODE;
     }
     // Write out temporarily. Will be overwritten at the end.
     saveHeader.segmentDescr = ftell(exports.exportFile);
@@ -649,15 +678,13 @@ void SaveRequest::Perform()
     (void)AddHierarchyEntry(fileName, saveHeader.timeStamp);
 
     delete[](descrs);
+
+    CheckMemory();
 }
 
 Handle SaveState(TaskData *taskData, Handle args)
 {
-    TCHAR fileNameBuff[MAXPATHLEN];
-    POLYUNSIGNED length =
-        Poly_string_to_C(DEREFHANDLE(args)->Get(0), fileNameBuff, MAXPATHLEN);
-    if (length > MAXPATHLEN)
-        raise_syscall(taskData, "File name too long", ENAMETOOLONG);
+    TempString fileNameBuff(Poly_string_to_T_alloc(DEREFHANDLE(args)->Get(0)));
     // The value of depth is zero for top-level save so we need to add one for hierarchy.
     unsigned newHierarchy = get_C_unsigned(taskData, DEREFHANDLE(args)->Get(1)) + 1;
 
@@ -694,7 +721,7 @@ public:
     // The fileName here is the last file loaded.  As well as using it
     // to load the name can also be printed out at the end to identify the
     // particular file in the hierarchy that failed.
-    TCHAR fileName[MAXPATHLEN];
+    AutoFree<TCHAR*> fileName;
     int errNumber;
 };
 
@@ -710,29 +737,47 @@ void StateLoader::Perform(void)
             return;
         }
         ML_Cons_Cell *p = DEREFLISTHANDLE(fileNameList);
-        POLYUNSIGNED length = Poly_string_to_C(p->h, fileName, MAXPATHLEN);
-        if (length > MAXPATHLEN)
+        fileName = Poly_string_to_T_alloc(p->h);
+        if (fileName == NULL)
         {
-            errorResult = "File name too long";
-            errNumber = ENAMETOOLONG;
+            errorResult = "Insufficient memory";
+            errNumber = NOMEMORY;
             return;
         }
         (void)LoadFile(true, 0, p->t);
     }
     else
     {
-        POLYUNSIGNED length =
-            Poly_string_to_C(DEREFHANDLE(fileNameList), fileName, MAXPATHLEN);
-        if (length > MAXPATHLEN)
+        fileName = Poly_string_to_T_alloc(DEREFHANDLE(fileNameList));
+        if (fileName == NULL)
         {
-            errorResult = "File name too long";
-            errNumber = ENAMETOOLONG;
+            errorResult = "Insufficient memory";
+            errNumber = NOMEMORY;
             return;
         }
         (void)LoadFile(true, 0, TAGGED(0));
     }
 }
 
+class ClearWeakByteRef: public ScanAddress
+{
+public:
+    ClearWeakByteRef() {}
+    virtual PolyObject *ScanObjectAddress(PolyObject *base) { return base; }
+    virtual void ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord);
+};
+
+// Set the values of external references and clear the values of other weak byte refs.
+void ClearWeakByteRef::ScanAddressesInObject(PolyObject *base, POLYUNSIGNED lengthWord)
+{
+    if (OBJ_IS_MUTABLE_OBJECT(lengthWord) && OBJ_IS_BYTE_OBJECT(lengthWord) && OBJ_IS_WEAKREF_OBJECT(lengthWord))
+    {
+        POLYUNSIGNED len = OBJ_OBJECT_LENGTH(lengthWord);
+        if (len > 0) base->Set(0, PolyWord::FromSigned(0));
+        setEntryPoint(base);
+    }
+}
+
 // This class is used to relocate addresses in areas that have been loaded.
 class LoadRelocate
 {
@@ -821,7 +866,7 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
     if ((FILE*)loadFile == NULL)
     {
         errorResult = "Cannot open load file";
-        errNumber = errno;
+        errNumber = ERRORNUMBER;
         return false;
     }
 
@@ -867,11 +912,11 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
                 return false;
             }
             ML_Cons_Cell *p = (ML_Cons_Cell *)tail.AsObjPtr();
-            POLYUNSIGNED length = Poly_string_to_C(p->h, fileName, MAXPATHLEN);
-            if (length > MAXPATHLEN)
+            fileName = Poly_string_to_T_alloc(p->h);
+            if (fileName == NULL)
             {
-                errorResult = "File name too long";
-                errNumber = ENAMETOOLONG;
+                errorResult = "Insufficient memory";
+                errNumber = NOMEMORY;
                 return false;
             }
             if (! LoadFile(false, header.parentTimeStamp, p->t))
@@ -880,7 +925,17 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
         else
         {
             size_t toRead = header.stringTableSize-header.parentNameEntry;
-            if (MAXPATHLEN < toRead) toRead = MAXPATHLEN;
+            size_t elems = ((toRead + sizeof(TCHAR) - 1) / sizeof(TCHAR));
+            // Always allow space for null terminator
+            size_t roundedBytes = (elems + 1) * sizeof(TCHAR);
+            TCHAR *newFileName = (TCHAR *)realloc(fileName, roundedBytes);
+            if (newFileName == NULL)
+            {
+                errorResult = "Insufficient memory";
+                errNumber = NOMEMORY;
+                return false;
+            }
+            fileName = newFileName;
 
             if (header.parentNameEntry >= header.stringTableSize /* Bad entry */ ||
                 fseek(loadFile, header.stringTable + header.parentNameEntry, SEEK_SET) != 0 ||
@@ -889,7 +944,7 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
                 errorResult = "Unable to read parent file name";
                 return false;
             }
-            fileName[toRead] = 0; // Should already be null-terminated, but just in case.
+            fileName[elems] = 0; // Should already be null-terminated, but just in case.
 
             if (! LoadFile(false, header.parentTimeStamp, TAGGED(0)))
                 return false;
@@ -952,8 +1007,7 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
     for (unsigned i = 0; i < relocate.nDescrs; i++)
     {
         SavedStateSegmentDescr *descr = &relocate.descrs[i];
-        MemSpace *space =
-            descr->segmentIndex == 0 ? gMem.IoSpace() : gMem.SpaceForIndex(descr->segmentIndex);
+        MemSpace *space = gMem.SpaceForIndex(descr->segmentIndex);
         if (space != NULL) relocate.targetAddresses[descr->segmentIndex] = space->bottom;
 
         if (descr->segmentData == 0)
@@ -974,9 +1028,9 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
             }
             // Allocate memory for the new segment.
             size_t actualSize = descr->segmentSize;
-            PolyWord *mem  =
-                (PolyWord*)osMemoryManager->Allocate(actualSize,
-                                PERMISSION_READ|PERMISSION_WRITE|PERMISSION_EXEC);
+            unsigned int perms = PERMISSION_READ|PERMISSION_WRITE;
+            if (descr->segmentFlags & SSF_CODE) perms |= PERMISSION_EXEC;
+            PolyWord *mem  = (PolyWord*)osMemoryManager->Allocate(actualSize, perms);
             if (mem == 0)
             {
                 errorResult = "Unable to allocate memory";
@@ -996,7 +1050,8 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
             unsigned mFlags =
                 (descr->segmentFlags & SSF_WRITABLE ? MTF_WRITEABLE : 0) |
                 (descr->segmentFlags & SSF_NOOVERWRITE ? MTF_NO_OVERWRITE : 0) |
-                (descr->segmentFlags & SSF_BYTES ? MTF_BYTES : 0);
+                (descr->segmentFlags & SSF_BYTES ? MTF_BYTES : 0) |
+                (descr->segmentFlags & SSF_CODE ? MTF_EXECUTABLE : 0);
             PermanentMemSpace *newSpace =
                 gMem.NewPermanentSpace(mem, actualSize / sizeof(PolyWord), mFlags,
                         descr->segmentIndex, hierarchyDepth+1);
@@ -1014,8 +1069,7 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
     for (unsigned j = 0; j < relocate.nDescrs; j++)
     {
         SavedStateSegmentDescr *descr = &relocate.descrs[j];
-        MemSpace *space =
-            descr->segmentIndex == 0 ? gMem.IoSpace() : gMem.SpaceForIndex(descr->segmentIndex);
+        MemSpace *space = gMem.SpaceForIndex(descr->segmentIndex);
         ASSERT(space != NULL); // We should have created it.
         if (descr->segmentFlags & SSF_OVERWRITE)
         {
@@ -1049,6 +1103,7 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
             if (fseek(loadFile, descr->relocations, SEEK_SET) != 0)
             {
                 errorResult = "Unable to read relocation segment";
+                return false;
             }
             for (unsigned k = 0; k < descr->relocationCount; k++)
             {
@@ -1056,9 +1111,9 @@ bool StateLoader::LoadFile(bool isInitial, time_t requiredStamp, PolyWord tail)
                 if (fread(&reloc, sizeof(reloc), 1, loadFile) != 1)
                 {
                     errorResult = "Unable to read relocation segment";
+                    return false;
                 }
-                MemSpace *toSpace =
-                    reloc.targetSegment == 0 ? gMem.IoSpace() : gMem.SpaceForIndex(reloc.targetSegment);
+                MemSpace *toSpace = gMem.SpaceForIndex(reloc.targetSegment);
                 if (toSpace == NULL)
                 {
                     errorResult = "Unknown space reference in relocation";
@@ -1097,11 +1152,11 @@ Handle LoadState(TaskData *taskData, bool isHierarchy, Handle hFileList)
             raise_fail(taskData, loader.errorResult);
         else
         {
-            char buff[MAXPATHLEN+100];
+            AutoFree<char*> buff((char *)malloc(strlen(loader.errorResult) + 2 + _tcslen(loader.fileName) * sizeof(TCHAR) + 1));
 #if (defined(_WIN32) && defined(UNICODE))
-            sprintf(buff, "%s: %S", loader.errorResult, loader.fileName);
+            sprintf(buff, "%s: %S", loader.errorResult, (TCHAR *)loader.fileName);
 #else
-            sprintf(buff, "%s: %s", loader.errorResult, loader.fileName);
+            sprintf(buff, "%s: %s", loader.errorResult, (TCHAR *)loader.fileName);
 #endif
             raise_syscall(taskData, buff, loader.errNumber);
         }
@@ -1139,28 +1194,25 @@ Handle ShowHierarchy(TaskData *taskData)
 Handle RenameParent(TaskData *taskData, Handle args)
 // Change the name of the immediate parent stored in a child
 {
-    TCHAR fileNameBuff[MAXPATHLEN], parentNameBuff[MAXPATHLEN];
     // The name of the file to modify.
-    POLYUNSIGNED fileLength =
-        Poly_string_to_C(DEREFHANDLE(args)->Get(0), fileNameBuff, MAXPATHLEN);
-    if (fileLength > MAXPATHLEN)
-        raise_syscall(taskData, "File name too long", ENAMETOOLONG);
+    AutoFree<TCHAR*> fileNameBuff(Poly_string_to_T_alloc(DEREFHANDLE(args)->Get(0)));
+    if (fileNameBuff == NULL)
+        raise_syscall(taskData, "Insufficient memory", NOMEMORY);
     // The new parent name to insert.
-    POLYUNSIGNED parentLength =
-        Poly_string_to_C(DEREFHANDLE(args)->Get(1), parentNameBuff, MAXPATHLEN);
-    if (parentLength > MAXPATHLEN)
-        raise_syscall(taskData, "Parent name too long", ENAMETOOLONG);
+    AutoFree<TCHAR*> parentNameBuff(Poly_string_to_T_alloc(DEREFHANDLE(args)->Get(1)));
+    if (parentNameBuff == NULL)
+        raise_syscall(taskData, "Insufficient memory", NOMEMORY);
 
     AutoClose loadFile(_tfopen(fileNameBuff, _T("r+b"))); // Open for reading and writing
     if ((FILE*)loadFile == NULL)
     {
-        char buff[MAXPATHLEN+1+23];
+        AutoFree<char*> buff((char *)malloc(23 + _tcslen(fileNameBuff) * sizeof(TCHAR) + 1));
 #if (defined(_WIN32) && defined(UNICODE))
-        sprintf(buff, "Cannot open load file: %S", fileNameBuff);
+        sprintf(buff, "Cannot open load file: %S", (TCHAR *)fileNameBuff);
 #else
-        sprintf(buff, "Cannot open load file: %s", fileNameBuff);
+        sprintf(buff, "Cannot open load file: %s", (TCHAR *)fileNameBuff);
 #endif
-        raise_syscall(taskData, buff, errno);
+        raise_syscall(taskData, buff, ERRORNUMBER);
     }
 
     SavedStateHeader header;
@@ -1203,22 +1255,20 @@ Handle RenameParent(TaskData *taskData, Handle args)
 Handle ShowParent(TaskData *taskData, Handle hFileName)
 // Return the name of the immediate parent stored in a child
 {
-    TCHAR fileNameBuff[MAXPATHLEN+1];
-    POLYUNSIGNED length =
-        Poly_string_to_C(DEREFHANDLE(hFileName), fileNameBuff, MAXPATHLEN);
-    if (length > MAXPATHLEN)
-        raise_syscall(taskData, "File name too long", ENAMETOOLONG);
+    AutoFree<TCHAR*> fileNameBuff(Poly_string_to_T_alloc(DEREFHANDLE(hFileName)));
+    if (fileNameBuff == NULL)
+        raise_syscall(taskData, "Insufficient memory", NOMEMORY);
 
     AutoClose loadFile(_tfopen(fileNameBuff, _T("rb")));
     if ((FILE*)loadFile == NULL)
     {
-        char buff[MAXPATHLEN+1+23];
+        AutoFree<char*> buff((char *)malloc(23 + _tcslen(fileNameBuff) * sizeof(TCHAR) + 1));
 #if (defined(_WIN32) && defined(UNICODE))
-        sprintf(buff, "Cannot open load file: %S", fileNameBuff);
+        sprintf(buff, "Cannot open load file: %S", (TCHAR *)fileNameBuff);
 #else
-        sprintf(buff, "Cannot open load file: %s", fileNameBuff);
+        sprintf(buff, "Cannot open load file: %s", (TCHAR *)fileNameBuff);
 #endif
-        raise_syscall(taskData, buff, errno);
+        raise_syscall(taskData, buff, ERRORNUMBER);
     }
 
     SavedStateHeader header;
@@ -1239,9 +1289,13 @@ Handle ShowParent(TaskData *taskData, Handle hFileName)
     // Does this have a parent?
     if (header.parentNameEntry != 0)
     {
-        TCHAR parentFileName[MAXPATHLEN+1];
         size_t toRead = header.stringTableSize-header.parentNameEntry;
-        if (MAXPATHLEN < toRead) toRead = MAXPATHLEN;
+        size_t elems = ((toRead + sizeof(TCHAR) - 1) / sizeof(TCHAR));
+        // Always allow space for null terminator
+        size_t roundedBytes = (elems + 1) * sizeof(TCHAR);
+        AutoFree<TCHAR*> parentFileName((TCHAR *)malloc(roundedBytes));
+        if (parentFileName == NULL)
+            raise_syscall(taskData, "Insufficient memory", NOMEMORY);
 
         if (header.parentNameEntry >= header.stringTableSize /* Bad entry */ ||
             fseek(loadFile, header.stringTable + header.parentNameEntry, SEEK_SET) != 0 ||
@@ -1249,7 +1303,7 @@ Handle ShowParent(TaskData *taskData, Handle hFileName)
         {
             raise_fail(taskData, "Unable to read parent file name");
         }
-        parentFileName[toRead] = 0; // Should already be null-terminated, but just in case.
+        parentFileName[elems] = 0; // Should already be null-terminated, but just in case.
         // Convert the name into a Poly string and then build a "Some" value.
         // It's possible, although silly, to have the empty string as a parent name.
         Handle resVal = SAVE(C_string_to_Poly(taskData, parentFileName));
@@ -1262,7 +1316,7 @@ Handle ShowParent(TaskData *taskData, Handle hFileName)
 
 // Module system
 #define MODULESIGNATURE "POLYMODU"
-#define MODULEVERSION   1
+#define MODULEVERSION   2
 
 typedef struct _moduleHeader
 {
@@ -1276,10 +1330,7 @@ typedef struct _moduleHeader
     // These entries contain the real data.
     off_t       segmentDescr;           // Position of segment descriptor table
     unsigned    segmentDescrCount;      // Number of segment descriptors in the table
-    off_t       stringTable;            // Pointer to the string table (zero if none)
-    size_t      stringTableSize;        // Size of string table
     time_t      timeStamp;              // The time stamp for this file.
-    uintptr_t   fileSignature;          // The signature for this file.
     time_t      executableTimeStamp;    // The time stamp for the parent executable.
     // Root
     uintptr_t   rootSegment;
@@ -1319,13 +1370,20 @@ void ModuleStorer::Perform()
     if (exporter.exportFile == NULL)
     {
         errorMessage = "Cannot open export file";
-        errCode = errno;
+        errCode = ERRORNUMBER;
         return;
     }
     // RunExport copies everything reachable from the root, except data from
     // the executable because we've set the hierarchy to 1, using CopyScan.
     // It builds the tables in the export data structure then calls exportStore
     // to actually write the data.
+    if (! root->Word().IsDataPtr())
+    {
+        // If we have a completely empty module the list may be null.
+        // This needs to be dealt with at a higher level.
+        errorMessage = "Module root is not an address";
+        return;
+    }
     exporter.RunExport(root->WordP());
     errorMessage = exporter.errorMessage; // This will be null unless there's been an error.
 }
@@ -1348,7 +1406,7 @@ void ModuleExport::exportStore(void)
         modHeader.rootSegment = mt->mtIndex;
         modHeader.rootOffset = (char*)this->rootFunction - (char*)mt->mtAddr;
     }
-    modHeader.timeStamp = time(NULL);
+    modHeader.timeStamp = getBuildTime();
     modHeader.segmentDescrCount = this->memTableEntries; // One segment for each space.
     // Write out the header.
     fwrite(&modHeader, sizeof(modHeader), 1, this->exportFile);
@@ -1375,6 +1433,8 @@ void ModuleExport::exportStore(void)
             if (entry->mtFlags & MTF_BYTES)
                 thisDescr->segmentFlags |= SSF_BYTES;
         }
+        if (entry->mtFlags & MTF_EXECUTABLE)
+             thisDescr->segmentFlags |= SSF_CODE;
     }
     // Write out temporarily. Will be overwritten at the end.
     modHeader.segmentDescr = ftell(this->exportFile);
@@ -1456,7 +1516,7 @@ void ModuleLoader::Perform()
     if ((FILE*)loadFile == NULL)
     {
         errorResult = "Cannot open load file";
-        errNumber = errno;
+        errNumber = ERRORNUMBER;
         return;
     }
 
@@ -1510,8 +1570,7 @@ void ModuleLoader::Perform()
     for (unsigned i = 0; i < relocate.nDescrs; i++)
     {
         SavedStateSegmentDescr *descr = &relocate.descrs[i];
-        MemSpace *space =
-            descr->segmentIndex == 0 ? gMem.IoSpace() : gMem.SpaceForIndex(descr->segmentIndex);
+        MemSpace *space = gMem.SpaceForIndex(descr->segmentIndex);
 
         if (descr->segmentData == 0)
         { // No data - just an entry in the index.
@@ -1532,11 +1591,30 @@ void ModuleLoader::Perform()
             }
             // Allocate memory for the new segment.
             size_t actualSize = descr->segmentSize;
-            LocalMemSpace *space = gMem.NewLocalSpace(actualSize, descr->segmentFlags & SSF_WRITABLE);
-            if (space == 0)
+            MemSpace *space;
+            if (descr->segmentFlags & SSF_CODE)
             {
-                errorResult = "Unable to allocate memory";
-                return;
+                CodeSpace *cSpace = gMem.NewCodeSpace(actualSize);
+                if (cSpace == 0)
+                {
+                    errorResult = "Unable to allocate memory";
+                    return;
+                }
+                space = cSpace;
+                cSpace->firstFree = (PolyWord*)((byte*)space->bottom + descr->segmentSize);
+                if (cSpace->firstFree != cSpace->top)
+                    gMem.FillUnusedSpace(cSpace->firstFree, cSpace->top - cSpace->firstFree);
+            }
+            else
+            {
+                LocalMemSpace *lSpace = gMem.NewLocalSpace(actualSize, descr->segmentFlags & SSF_WRITABLE);
+                if (lSpace == 0)
+                {
+                    errorResult = "Unable to allocate memory";
+                    return;
+                }
+                space = lSpace;
+                lSpace->lowerAllocPtr = (PolyWord*)((byte*)lSpace->bottom + descr->segmentSize);
             }
             if (fseek(loadFile, descr->segmentData, SEEK_SET) != 0 ||
                 fread(space->bottom, descr->segmentSize, 1, loadFile) != 1)
@@ -1544,12 +1622,11 @@ void ModuleLoader::Perform()
                 errorResult = "Unable to read segment";
                 return;
             }
-            space->lowerAllocPtr = (PolyWord*)((byte*)space->bottom + descr->segmentSize);
             relocate.targetAddresses[descr->segmentIndex] = space->bottom;
             if (space->isMutable && (descr->segmentFlags & SSF_BYTES) != 0)
             {
                 ClearWeakByteRef cwbr;
-                cwbr.ScanAddressesInRegion(space->bottom, space->lowerAllocPtr);
+                cwbr.ScanAddressesInRegion(space->bottom, (PolyWord*)((byte*)space->bottom + descr->segmentSize));
             }
         }
     }
@@ -1599,7 +1676,7 @@ Handle LoadModule(TaskData *taskData, Handle args)
             raise_fail(taskData, loader.errorResult);
         else
         {
-            char buff[MAXPATHLEN+100];
+            AutoFree<char*> buff((char *)malloc(strlen(loader.errorResult) + 2 + _tcslen(loader.fileName) * sizeof(TCHAR) + 1));
 #if (defined(_WIN32) && defined(UNICODE))
             sprintf(buff, "%s: %S", loader.errorResult, loader.fileName);
 #else
