@@ -38,7 +38,6 @@
 #include "globals.h"
 #include "scanaddrs.h"
 #include "machine_dep.h"
-#include "check_objects.h"
 #include "diagnostics.h"
 #include "memmgr.h"
 
@@ -52,26 +51,10 @@ POLYUNSIGNED ScanAddress::ScanAddressAt(PolyWord *pt)
         // We can get zeros in the constant area if we garbage collect
         //  while compiling some code. */
     }
-    else if (val.IsCodePtr())
-    {
-        // We can get code pointers either in the stack as return addresses or
-        // handler pointers or in constants in code segments as the addresses of
-        // exception handlers.
-
-        // Find the start of the code segment
-        PolyObject *oldObject = ObjCodePtrToPtr(val.AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = val.AsCodePtr() - (byte*)oldObject;
-        // Mustn't use ScanAddressAt here.  That's only valid if the value points
-        // into the area being updated.
-        PolyObject *newObject = ScanObjectAddress(oldObject);
-        newVal = PolyWord::FromCodePtr((byte*)newObject + offset);
-    }
     else
     {
         ASSERT(OBJ_IS_DATAPTR(val));
-        // Database pointer, local pointer or IO pointer.
-        // We need to include IO area pointers when we produce an object module.
+        // Any sort of address
         newVal = ScanObjectAddress(val.AsObjPtr());
     }
     if (newVal != val) // Only update if we need to.
@@ -79,40 +62,6 @@ POLYUNSIGNED ScanAddress::ScanAddressAt(PolyWord *pt)
     return 0;
 }
 
-// Process a value within the stack.
-PolyWord ScanAddress::ScanStackAddress(PolyWord val, StackSpace *stack, bool isCode)
-{
-    PolyWord *base = stack->bottom;
-    PolyWord *end = stack->top;
-
-    // If isCode is set we definitely have a code address.  It may have the
-    // bottom bit set or it may be word aligned.
-    if (isCode || val.IsCodePtr())
-    {
-        /* Find the start of the code segment */
-        PolyObject *oldObject = ObjCodePtrToPtr(val.AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = val.AsCodePtr() - (byte*)oldObject;
-        PolyObject *newObject = ScanObjectAddress(oldObject);
-        return PolyWord::FromCodePtr((byte*)newObject + offset);
-    }
-
-    else if (val.IsTagged() || val == PolyWord::FromUnsigned(0) || 
-                 (val.AsAddress() > base && val.AsAddress() <= end))
-            /* We don't need to process tagged integers (now we've checked it isn't
-               a code address) and we don't need to process addresses within the
-               current stack. */
-            /* N.B. We have "<= end" rather than "< end" because it is possible for
-               the stack to be completely empty on a terminated thread. */
-           return val;
-
-    else
-    {
-        ASSERT(val.IsDataPtr());
-        return ScanObjectAddress(val.AsObjPtr());
-    }
-}
-
 // General purpose object processor,  Processes all the addresses in an object.
 // Handles the various kinds of object that may contain addresses.
 void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord)
@@ -175,10 +124,8 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
                 {
                     wordAt = *baseAddr; // Reload because it may have been side-effected
                      // We really have to process this recursively.
-                    if (wordAt.IsCodePtr())
-                        ScanAddressesInObject(ObjCodePtrToPtr(wordAt.AsCodePtr()), lengthWord);
-                    else
-                        ScanAddressesInObject(wordAt.AsObjPtr(), lengthWord);
+                    ASSERT(wordAt.IsDataPtr());
+                    ScanAddressesInObject(wordAt.AsObjPtr(), lengthWord);
                     baseAddr++;
                 }
                 else baseAddr++;
@@ -189,10 +136,8 @@ void ScanAddress::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord
         // Do this by looping rather than recursion.
         PolyWord wordAt = *baseAddr; // Last word to do.
         // This must be an address 
-        if (wordAt.IsCodePtr())
-            obj = ObjCodePtrToPtr(wordAt.AsCodePtr());
-        else
-            obj = wordAt.AsObjPtr();
+        ASSERT(wordAt.IsDataPtr());
+        obj = wordAt.AsObjPtr();
 
         lengthWord = lastLengthWord;
 
@@ -213,10 +158,8 @@ void ScanAddress::ScanAddressesInRegion(PolyWord *region, PolyWord *end)
             // of applying ShareData repeatedly.  Perhaps we should
             // turn the forwarding pointers back into normal words in
             // an extra pass.
-            while (obj->ContainsForwardingPtr())
-                obj = obj->GetForwardingPtr();
+            obj = obj->FollowForwardingChain();
             ASSERT(obj->ContainsNormalLengthWord());
-            CheckObject(obj);
             pt += obj->Length();
         }
         else
@@ -246,13 +189,6 @@ PolyWord ScanAddress::GetConstantValue(byte *addressOfConstant, ScanRelocationKi
             for (i = sizeof(PolyWord); i > 0; i--)
                 valu = (valu << 8) | pt[i-1];
 
-            /* The old code generator generated reverse subtraction
-               of words using a move immediate which loaded a register
-               with a the tagged value plus one.  In practice the only
-               reverse subtraction of a constant is 0-x so for backwards
-               compatibility we need to treat 2 specially. */
-            ASSERT(valu != 2);
-
             return PolyWord::FromUnsigned(valu);
         }
     case PROCESS_RELOC_I386RELATIVE:         // 32 bit relative address
@@ -307,7 +243,7 @@ void ScanAddress::SetConstantValue(byte *addressOfConstant, PolyWord p, ScanRelo
 
 // The default action is to call the DEFAULT ScanAddressAt NOT the virtual which means that it calls
 // ScanObjectAddress for the base address of the object referred to.
-void ScanAddress::ScanConstant(byte *addressOfConstant, ScanRelocationKind code)
+void ScanAddress::ScanConstant(PolyObject *base, byte *addressOfConstant, ScanRelocationKind code)
 {
     PolyWord p = GetConstantValue(addressOfConstant, code);
 
@@ -323,17 +259,6 @@ void ScanAddress::ScanConstant(byte *addressOfConstant, ScanRelocationKind code)
 void ScanAddress::ScanRuntimeWord(PolyWord *w)
 {
     if (w->IsTagged()) {} // Don't need to do anything
-    else if (w->IsCodePtr())
-    {
-        // We can have code pointers in set_code_address.
-        // Find the start of the code segment
-        PolyObject *obj = ObjCodePtrToPtr(w->AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = w->AsCodePtr() - (byte*)obj;
-        obj = ScanObjectAddress(obj); 
-        *w = PolyWord::FromCodePtr((byte*)obj + offset);
-
-    }
     else {
         ASSERT(w->IsDataPtr());
         *w = ScanObjectAddress(w->AsObjPtr()); 
@@ -376,10 +301,7 @@ PolyObject *RecursiveScan::ScanObjectAddress(PolyObject *obj)
 void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWord)
 {
     if (OBJ_IS_BYTE_OBJECT(lengthWord))
-    {
-        Completed(obj);
-        return;
-    }
+        return; // Ignore byte cells and don't call Completed on them
 
     while (true)
     {
@@ -396,7 +318,6 @@ void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWo
             ScanAddress::ScanAddressesInObject(obj, lengthWord);
             length = 0; // Finished
         }
-        ASSERT(! OBJ_IS_BYTE_OBJECT(lengthWord)); // Check - remove this later
 
         // else it's a normal object,
 
@@ -441,32 +362,6 @@ void RecursiveScan::ScanAddressesInObject(PolyObject *obj, POLYUNSIGNED lengthWo
                     else break;  // More than two words.
                 }
             }
-            else if (wordAt.IsCodePtr())
-            {
-                // If we're processing the constant area of a code segment this could
-                // be a code address.
-                PolyObject *oldObject = ObjCodePtrToPtr(wordAt.AsCodePtr());
-                // Calculate the byte offset of this value within the code object.
-                POLYUNSIGNED offset = wordAt.AsCodePtr() - (byte*)oldObject;
-                wordAt = oldObject;
-                bool test = TestForScan(&wordAt);
-                // TestForScan may side-effect the word.
-                PolyObject *newObject = wordAt.AsObjPtr();
-                wordAt = PolyWord::FromCodePtr((byte*)newObject + offset);
-                if (wordAt != *baseAddr)
-                    *baseAddr = wordAt;
-                if (test)
-                {
-                    if (firstWord == 0)
-                    {
-                        firstWord = newObject;
-                        MarkAsScanning(firstWord);
-                    }
-                    else if (secondWord == 0)
-                        secondWord = newObject;
-                    else break;
-                }
-            }
             baseAddr++;
         }
 
