@@ -64,7 +64,7 @@ public:
     // Process a constant within the code.
     // The default action is to call the DEFAULT ScanAddressAt NOT the virtual which means that it calls
     // ScanObjectAddress for the base address of the object referred to.
-    virtual void ScanConstant(byte *addressOfConstant, ScanRelocationKind code);
+    virtual void ScanConstant(PolyObject *base, byte *addressOfConstant, ScanRelocationKind code);
 
     // Scan the objects in the region and process their addresses.  Applies ScanAddressesInObject
     // to each of the objects.  The "region" argument points AT the first length word.
@@ -86,13 +86,6 @@ protected:
     void ScanAddressesInObject(PolyObject *base) { ScanAddressesInObject(base, base->LengthWord()); }
 
 public:
-    // Process a word found on the stack and return the (updated) value.  Stacks may contain
-    // code addresses and addresses within the current stack.  "isCode" is set to true for
-    // the program counter which may be a code address but not satisfy IsCodeObject.
-    // The default action is to ignore integers and addresses within the stack and to call
-    // ScanObjectAddress for the base addresses of all other addresses.
-    PolyWord ScanStackAddress(PolyWord val, StackSpace *base, bool isCode);
-
     // Extract a constant from the code.
     static PolyWord GetConstantValue(byte *addressOfConstant, ScanRelocationKind code);
     // Store a constant in the code.
