@@ -3,7 +3,7 @@
 
     Copyright (c) 2000
         Cambridge University Technical Services Limited
-    and David C. J. Matthews 2006, 2010-13
+    and David C. J. Matthews 2006, 2010-13, 2016
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -51,14 +51,15 @@
 #include "run_time.h"
 #include "sys.h"
 #include "gc.h"
-#include "rts_module.h"
+#include "rtsentry.h"
 #include "memmgr.h"
 #include "processes.h"
 #include "gctaskfarm.h"
 #include "diagnostics.h"
+#include "sharedata.h"
 
 /*
-This code was largely written by Simon Finn as a database improver for the the
+This code was largely written by Simon Finn as a database improver for the
 memory-mapped persistent store version.  The aim is that where two immutable
 objects (cells) contain the same data (i.e. where ML equality would say they
 were equal) they should be merged so that only a single object is retained.
@@ -112,6 +113,9 @@ now in dynamic memory.  That avoids a possible segfault if the normal
 C stack overflows.
 */
 
+extern "C" {
+    POLYEXTERNALSYMBOL POLYUNSIGNED PolyShareCommonData(PolyObject *threadId, PolyWord root);
+}
 typedef struct
 {
     POLYUNSIGNED    L;
@@ -159,8 +163,8 @@ private:
 ShareDataClass::~ShareDataClass()
 {
     // Free the bitmaps associated with the permanent spaces.
-    for (unsigned j = 0; j < gMem.npSpaces; j++)
-        gMem.pSpaces[j]->shareBitmap.Destroy();
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
+       (*i)->shareBitmap.Destroy();
 }
 
 DepthVector *ShareDataClass::AddDepth(POLYUNSIGNED depth)
@@ -279,7 +283,7 @@ POLYUNSIGNED DepthVector::MergeSameItems()
             // If we can't find a permanent space choose a space that isn't
             // an allocation space.  Otherwise we could break the invariant
             // that immutable areas never point into the allocation area.
-            MemSpace *space = gMem.SpaceForAddress(itemVec[j].pt);
+            MemSpace *space = gMem.SpaceForAddress(itemVec[j].pt-1);
             if (bestSpace == 0)
             {
                 bestShare = itemVec[j].pt;
@@ -450,21 +454,9 @@ POLYUNSIGNED ProcessFixupAddress::ScanAddressAt(PolyWord *pt)
 // has moved, otherwise returns the original.
 PolyWord ProcessFixupAddress::GetNewAddress(PolyWord old)
 {
-    if (old.IsTagged() || old == PolyWord::FromUnsigned(0) || gMem.IsIOPointer(old.AsAddress()))
+    if (old.IsTagged() || old == PolyWord::FromUnsigned(0))
         return old; //  Nothing to do.
 
-    // When we are updating addresses in the stack or in code segments we may have
-    // code pointers.
-    if (old.IsCodePtr())
-    {
-        // Find the start of the code segment
-        PolyObject *oldObject = ObjCodePtrToPtr(old.AsCodePtr());
-        // Calculate the byte offset of this value within the code object.
-        POLYUNSIGNED offset = old.AsCodePtr() - (byte*)oldObject;
-        PolyWord newObject = GetNewAddress(oldObject);
-        return PolyWord::FromCodePtr(newObject.AsCodePtr() + offset);
-    }
-
     ASSERT(old.IsDataPtr());
 
     PolyObject *obj = old.AsObjPtr();
@@ -551,8 +543,8 @@ POLYUNSIGNED ProcessAddToVector::AddObjectsToDepthVectors(PolyWord old)
     if (old.IsTagged() || old == PolyWord::FromUnsigned(0))
         return 0;
 
-    MemSpace *space = gMem.SpaceForAddress(old.AsAddress());
-    if (space == 0 || space->spaceType == ST_IO)
+    MemSpace *space = gMem.SpaceForAddress(old.AsStackAddr()-1);
+    if (space == 0)
         return 0;
 
     PolyObject *obj = old.AsObjPtr();
@@ -759,9 +751,9 @@ bool ShareDataClass::RunShareData(PolyObject *root)
 {
     // We use a bitmap to indicate when we've visited an object to avoid
     // infinite recursion in cycles in the data.
-    for (unsigned j = 0; j < gMem.npSpaces; j++)
+    for (std::vector<PermanentMemSpace*>::iterator i = gMem.pSpaces.begin(); i < gMem.pSpaces.end(); i++)
     {
-        PermanentMemSpace *space = gMem.pSpaces[j];
+        PermanentMemSpace *space = *i;
         if (!space->isMutable && space->hierarchy == 0)
         {
             if (! space->shareBitmap.Create(space->spaceSize()))
@@ -903,3 +895,37 @@ void ShareData(TaskData *taskData, Handle root)
     if (! request.result)
         raise_exception_string(taskData, EXC_Fail, "Insufficient memory");
 }
+
+// RTS call entry.
+POLYUNSIGNED PolyShareCommonData(PolyObject *threadId, PolyWord root)
+{
+    TaskData *taskData = TaskData::FindTaskForId(threadId);
+    ASSERT(taskData != 0);
+    taskData->PreRTSCall();
+    Handle reset = taskData->saveVec.mark();
+
+    try {
+        if (! root.IsDataPtr())
+            return TAGGED(0).AsUnsigned(); // Nothing to do.
+
+        // Request the main thread to do the sharing.
+        ShareRequest request(taskData->saveVec.push(root));
+        processes->MakeRootRequest(taskData, &request);
+
+        // Raise an exception if it failed.
+        if (! request.result)
+            raise_exception_string(taskData, EXC_Fail, "Insufficient memory");
+    } catch (...) { } // If an ML exception is raised
+
+    taskData->saveVec.reset(reset);
+    taskData->PostRTSCall();
+    return TAGGED(0).AsUnsigned();
+}
+
+struct _entrypts shareDataEPT[] =
+{
+    { "PolyShareCommonData",            (polyRTSFunction)&PolyShareCommonData},
+
+    { NULL, NULL} // End of list.
+};
+
